from .base import BaseSamplingElement
import numpy as np


class MCMCSampler(BaseSamplingElement, sampler_name='mcmc_sampler'):
    """A Metropolis-Hastings MCMC Sampler.

    Parameters
    ----------
    init: dict
       Initial values for each input parameter. Of the form {'input1': value, ...}
    q: function
       A function of one argument X (dictionary) that returns the proposal distribution conditional on
    the X.
    qoi: str
       Name of the quantity of interest
    n_chains: int
       Number of MCMC chains to run in paralle.
    estimator: function
       To be used with replica_col argument. Outputs an estimate of some
       parameter when given a sample array.
    """

    def __init__(self, init, q, qoi, n_chains=1, likelihood=lambda x: x[0], estimator=None):
        self.init = dict(init)
        self.inputs = list(self.init.keys())
        for input_ in self.inputs:
            if len(self.init[input_]) != n_chains:
                raise RuntimeError("The init dictionary must contains the same number \
                                    of values for each input as there are chains.")
        self.n_chains = n_chains
        self.x = []
        self.q = q
        self.qoi = qoi
        self.current_chain = 0
        for chain in range(self.n_chains):
            self.x.append(dict([(key, self.init[key][chain]) for key in self.inputs]))
            self.x[chain]['chain_id'] = chain
        self.f_x = [None] * n_chains
        self.stop = False
        self.likelihood = lambda x: np.exp(likelihood(x))
        self.n_replicas = None
        self.estimator = estimator
        self.acceptance_ratios = []
        self.iteration = 0

    def is_finite(self):
        return True

    def n_samples(self):
        return self.n_chains

    def __iter__(self):
        self.current_chain = 0
        return self

    def __next__(self):
        """Returns next MCMC sample.

        Returns
        -------
        dict
           A dictionary where keys are input variables names and values are input values.
        """
        if self.stop:
            raise StopIteration
        if self.f_x[self.current_chain] is None:
            try:
                return self.x[self.current_chain]
            finally:
                self.current_chain = (self.current_chain + 1) % self.n_chains
                if self.current_chain == 0:
                    self.stop = True
        y = {}
        y_ = self.q(self.x[self.current_chain]).sample()
        for i, key in enumerate(self.inputs):
            y[key] = y_[i][0]
        y['chain_id'] = self.current_chain
        self.current_chain = (self.current_chain + 1) % self.n_chains
        if self.current_chain == 0:
            self.stop = True
        return y

    def update(self, result, invalid):
        """Performs the MCMC sampling procedure on the campaign.

        Parameters
        ----------
        result: pandas DataFrame
            run information from previous iteration (same as collation DataFrame)
        invalid: pandas DataFrame
            invalid run information (runs that cannot be executed for some reason)

        Returns
        -------
        list of rejected run ids
        """
        self.stop = False
        if (self.estimator is not None) and (len(result) > 0):
            result_grouped = result.groupby(('chain_id', 0)).apply(self.estimator)
        else:
            result_grouped = result
        if (self.estimator is not None) and (len(invalid) > 0):
            invalid_grouped = invalid.groupby(('chain_id', 0)).apply(lambda x: x.mean())
        else:
            invalid_grouped = invalid
        ignored_chains = []
        ignored_runs = []
        # process normal runs
        for row in result_grouped.iterrows():
            row = row[1]
            chain_id = int(row['chain_id'].values[0])
            y = dict([(key, row[key][0]) for key in self.inputs])
            if self.f_x[chain_id] is None:
                self.f_x[chain_id] = self.likelihood(row[self.qoi].values)
            else:
                f_y = self.likelihood(row[self.qoi].values)
                q_xy = self.q(y).pdf([self.x[chain_id][key] for key in self.inputs])
                q_yx = self.q(self.x[chain_id]).pdf([y[key] for key in self.inputs])
                if self.f_x[chain_id] == 0.0:
                    r = 1.0
                else:
                    r = min(1.0, (f_y / self.f_x[chain_id]) * (q_xy / q_yx))
                if np.random.random() < r:
                    self.x[chain_id] = dict(y)
                    self.f_x[chain_id] = f_y
                else:
                    ignored_chains.append(chain_id)
        for row in invalid_grouped.iterrows():
            row = row[1]
            chain_id = int(row['chain_id'].values[0])
            ignored_chains.append(chain_id)
        for chain_id in ignored_chains:
            try:
                ignored_runs += list(result.loc[result[('chain_id', 0)]
                                                == chain_id]['run_id'].values)
            except KeyError:
                pass
            try:
                ignored_runs += list(invalid.loc[invalid[('chain_id', 0)]
                                                 == chain_id]['run_id'].values)
            except KeyError:
                pass
        ignored_runs = [run[0] for run in ignored_runs]
        self.iteration += 1
        return ignored_runs

    @property
    def analysis_class(self):
        """Returns a corresponding analysis class for this sampler.

        Returns
        -------
        class
        """
        from easyvvuq.analysis import MCMCAnalysis
        return MCMCAnalysis



import easyvvuq as uq
import chaospy as cp
import numpy as np
import json
from timeit import Timer

def benchmark(nsamples):
    params = {
        "S0": {"type": "float", "default": 997},
        "I0": {"type": "float", "default": 3},
        "beta": {"type": "float", "default": 0.2},
        "gamma": {"type": "float", "default": 0.04, "min": 0.0, "max": 1.0},
        "iterations": {"type": "integer", "default": 100},
        "outfile": {"type": "string", "default": "output.csv"}
    }
    encoder = uq.encoders.GenericEncoder(
        template_fname='tutorials/sir.template',
        delimiter='$',
        target_filename='input.json')
    decoder = uq.decoders.SimpleCSV(target_filename='output.csv', output_columns=['I'])
    actions = uq.actions.local_execute(encoder, 'sir input.json', decoder)
    campaign = uq.Campaign(name='sir_benchmark', params=params, actions=actions)
    vary = {
        "beta": cp.Uniform(0.15, 0.25),
        "gamma": cp.Normal(0.04, 0.001),
    }
    campaign.set_sampler(uq.sampling.RandomSampler(vary=vary))
    benchmark_results = {}
    t = Timer('campaign.draw_samples(nsamples)', globals=locals())
    benchmark_results['draw_samples'] = t.timeit(1)
    def fake_results():
        counter = 1
        while True:
            yield ('Run_{}'.format(counter), {'values': list(np.random.random(size=100))})
            counter += 1
    fakes = fake_results()
    results_ = []
    for _ in range(nsamples):
        results_.append(next(fakes))
    t = Timer('campaign.campaign_db.store_results(campaign._active_app_name, results_)', globals=locals())
    benchmark_results['store_results'] = t.timeit(1)
    t = Timer('campaign.get_collation_result()', globals=locals())
    benchmark_results['get_collation_result'] = t.timeit(1)
    return benchmark_results
    
    

if __name__ == '__main__':
    nsamples = [1000, 10000, 100000, 1000000]
    results = {}
    for nsamples_ in nsamples:
        results[nsamples_] = benchmark(nsamples_)
    print(json.dumps(results))



#!/usr/bin/env python3

import json
import sys
import numpy as np
import math

# Author: Wouter Edeling

# solves the steady advection diffusion equation at Peclet number Pe and
# constant forcing term f, using Finite Elements with linear shape functions.


def solve(Pe, f, nel):

    print('Solving advection diffusion equation at Pe = ', Pe, ' and f = ', f)

    h = 1.0 / nel

    K = np.zeros([nel + 1, nel + 1])
    F = np.zeros(nel + 1)

    # GQ RULES
    Q = 2
    eps_i = np.array([-1 / math.sqrt(3.0), 1 / math.sqrt(3.0)])
    wi = np.ones(2)

    # assumble K matrix and F vector
    for e in range(nel):

        XA1 = e * h
        XA2 = (e + 1) * h

        x = (h * eps_i + XA1 + XA2) / 2

        [N1x, N2x] = der_shape(h, Q)
        [N1, N2] = shape(x, XA1, XA2)

        # diffusive term
        k11 = 1.0 / Pe * N1x * N1x * wi - N1x * N1 * wi
        k11 = k11.sum()
        k12 = 1.0 / Pe * N1x * N2x * wi - N1x * N2 * wi
        k12 = k12.sum()
        k21 = 1.0 / Pe * N2x * N1x * wi - N2x * N1 * wi
        k21 = k21.sum()
        k22 = 1.0 / Pe * N2x * N2x * wi - N2x * N2 * wi
        k22 = k22.sum()

        # forcing term
        f1 = N1 * f
        f1 = f1.sum()
        f2 = N2 * f
        f2 = f2.sum()

        # local element matrices
        if e > 0 and e < int(nel - 1):
            K[e, e] = K[e, e] + k11
            K[e, e + 1] = K[e, e + 1] + k12
            K[e + 1, e] = K[e + 1, e] + k21
            K[e + 1, e + 1] = K[e + 1, e + 1] + k22

            F[e] = F[e] + f1
            F[e + 1] = F[e + 1] + f2
        elif e == 0:
            K[e + 1, e] = K[e + 1, e] + k21
            K[e + 1, e + 1] = K[e + 1, e + 1] + k22

            F[e + 1] = F[e + 1] + f2
        else:
            K[e, e] = K[e, e] + k11
            K[e, e + 1] = K[e, e + 1] + k12

            F[e] = F[e] + f1

    K[0, 0] = 1.0
    K[nel, nel] = 1.0

    return np.linalg.solve(K, F)

# Finite element linear shape functions and their derivatives


def der_shape(h, Q):
    N1x = -1 / h * np.ones(Q)
    N2x = 1 / h * np.ones(Q)

    return [N1x, N2x]


def shape(x, XA1, XA2):

    if np.max(x) < 1.0:
        eps = (2.0 * x - XA1 - XA2) / (XA2 - XA1)
    else:
        eps = (2.0 * x - XA1 - XA2) / (XA2 - XA1)
        eps[-1] = -1

    # value shape function
    N1 = 0.5 * (1 - eps)
    N2 = 0.5 * (1 + eps)
    return [N1, N2]


# the json input file containing the values of the parameters, and the
# output file
json_input = sys.argv[1]

with open(json_input, "r") as f:
    inputs = json.load(f)

Pe = float(inputs['Pe'])
f = float(inputs['f'])
output_filename = inputs['outfile']

# run FEM solver at current value of Pe and f
u = solve(Pe, f, nel=300)
u[0] = 0.0

# output csv file
header = 'u'
np.savetxt(output_filename, u,
           delimiter=",", comments='',
           header=header)



"""Data structures to ensure consistency during serialization for databases.

"""
import os
import logging
import json
from easyvvuq import constants
from easyvvuq.utils.helpers import easyvvuq_serialize
import numpy

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)


def check_local_dir(path, dir_type='campaign'):
    """
    Check that local path exists and if not create it.

    Parameters
    ----------
    path : str
        Directory location to check.
    dir_type : str, default='campaign'
        Type of directory we are checking (used for user and debugging
        information.)

    Returns
    -------

    """

    if not os.path.isdir(path):

        if os.path.exists(path):
            logger.critical(f'{path} specified as {dir_type} directory '
                            f'for local run but is not a directory.')
            raise IOError(f'Invalid {dir_type} directory')
        else:
            os.makedirs(path)


def check_reference(ref, run_name, ref_type='campaign'):
    """
    Validation check for a `RunInfo` reference. Checks that an integer value
    has been passed to use as a reference to another 'table' - i.e. to a
    specific campaign, app or sampler.

    Parameters
    ----------
    ref : int
        Reference to be checked.
    run_name : str
        Name of run for which the check is being performed (user info/
        debugging).
    ref_type : str, default='campaign'
        Are we checking for a campaign, sampler or app (user info/
        debugging).

    Returns
    -------

    """

    if ref is None:
        message = f'No {ref_type} id specified for run {run_name}'
        logger.critical(message)
        raise RuntimeError(message)

    if not isinstance(ref, int):
        message = (f'Invalid {ref_type} id ({ref}) specified for '
                   f'run {run_name}')
        logger.critical(message)
        raise RuntimeError(message)


class RunInfo:
    """Handles information for individual application runs.

    Parameters
    ----------
    run_name : str
        Human readable name of the run.
    app : None or int
        ID of the associated application.
    params : None or dict
        Dictionary of parameter values for this run.
    sample: None or int
        ID of the sampler that created the run.
    campaign: None or int
        ID of the associated campaign.

    Attributes
    ----------
    campaign : int
        ID of the associated campaign.
    sample : int
        ID of the sampler that created the run.
    app : int
        ID of the associated application.
    run_name : str
        Human readable name of the run.
    status : enum(Status)
    """

    def __init__(
            self,
            run_name=None,
            run_dir=None,
            app=None,
            params=None,
            sample=None,
            campaign=None,
            status=constants.Status.NEW):

        self.campaign = campaign
        self.sample = sample
        self.app = app
        self.run_name = run_name
        self.run_dir = run_dir

        if not params:
            message = f'No run configuration specified for run {run_name}'
            raise RuntimeError(message)

        self.params = params
        self.status = status

        self.iteration = 0

    def to_dict(self, flatten=False):
        """Convert to a dictionary (optionally flatten to single level)

        Parameters
        ----------
        flatten : bool
            Should the return dictionary be single level (i.e. should `params`
            or other dictionary variables be serialized).

        Returns
        -------
        dict
            Dictionary representing the run - if flattened then params are
            returned as a JSON format sting.
        """

        def convert_nonserializable(obj):
            if isinstance(obj, numpy.int64):
                return int(obj)
            raise TypeError('Unknown type:', type(obj))

        if flatten:

            out_dict = {
                'run_name': self.run_name,
                'run_dir': self.run_dir,
                'params': json.dumps(self.params, default=convert_nonserializable),
                'status': constants.Status(self.status),
                'campaign': self.campaign,
                'sampler': self.sample,
                'app': self.app,
                'iteration': self.iteration,
            }

        else:

            out_dict = {
                'run_name': self.run_name,
                'run_dir': self.run_dir,
                'params': self.params,
                'status': constants.Status(self.status),
                'campaign': self.campaign,
                'sampler': self.sample,
                'app': self.app,
                'iteration': self.iteration,
            }

        return out_dict


class AppInfo:
    """Handles information for particular application.

    Attributes
    ----------
    name : str or None
        Human readable application name.
    paramsspec : ParamsSpecification or None
        Description of possible parameter values.
    """

    def __init__(
            self,
            name=None,
            paramsspec=None,
            actions=None):

        self.name = name
        self.paramsspec = paramsspec
        self.actions = actions

    def to_dict(self, flatten=False):
        """Convert to a dictionary (optionally flatten to single level)

        Parameters
        ----------
        flatten : bool
            Should the return dictionary be single level (i.e. should `paramsspec`
            be serialized).

        Returns
        -------
        dict
            Dictionary representing the application- if flattened then `paramsspec`
            is returned as a JSON format sting.
        """

        if flatten:

            out_dict = self.to_dict()

            out_dict['params'] = self.paramsspec.serialize()
        else:

            out_dict = {
                'name': self.name,
                'params': self.paramsspec,
                'actions': easyvvuq_serialize(self.actions),
            }

        return out_dict


class CampaignInfo:
    """Handles information on Campaign.

    Parameters
    ----------
    name : str or None
        Human readable campaign name.
    easyvvuq_version : str or None
        Version of EasyVVUQ used to create the campaign.
    campaign_dir_prefix : str or None
        Prefix test for campaign directory.
    campaign_dir : str or None,
        Path to the campaign directory.
    runs_dir : str or None
        path to run directory (within the campaign directory)
    local : bool, default=False
        Is this campaign designed to be created and executed on the same
        machine?

    Attributes
    ----------
    name : str or None
        Human readable campaign name.
    easyvvuq_version : str or None
        Version of EasyVVUQ used to create the campaign.
    campaign_dir_prefix : str or None
        Prefix test for campaign directory.
    campaign_dir : str or None,
        Path to the campaign directory.
    runs_dir : str or None
        path to run directory (within the campaign directory)
    """

    def __init__(self, name=None, easyvvuq_version=None,
                 campaign_dir_prefix=None, campaign_dir=None,
                 runs_dir=None, local=False):

        if name is None:
            message = "CampaignInfo constructor must be passed a 'name'."
            logger.critical(message)
            raise RuntimeError(message)

        if campaign_dir is None:
            message = "CampaignInfo constructor must be passed 'campaign_dir'"
            logger.critical(message)
            raise RuntimeError(message)

        self.name = name
        self.campaign_dir_prefix = campaign_dir_prefix

        self.easyvvuq_version = easyvvuq_version

        # TODO: think about right location for path check for remote runs
        if local:
            check_local_dir(campaign_dir)

        self.campaign_dir = campaign_dir

        if runs_dir is None:
            runs_dir = os.path.join(campaign_dir, 'runs')

        if local:
            check_local_dir(runs_dir, 'runs')

        self.runs_dir = runs_dir

    @property
    def easyvvuq_version(self):
        return self._easyvvuq_version

    @easyvvuq_version.setter
    def easyvvuq_version(self, version_no):
        # TODO: check validity and compatibility
        self._easyvvuq_version = version_no

    def to_dict(self, flatten=False):
        """Convert this to a dictionary

        Parameters
        ----------
        flatten : bool
            Should the return dictionary be single level (always true here).

        Returns
        -------
        dict
            Dictionary representing the campaign.
        """

        out_dict = {
            'name': self.name,
            'campaign_dir': self.campaign_dir,
            'campaign_dir_prefix': self.campaign_dir_prefix,
            'runs_dir': self.runs_dir,
            'easyvvuq_version': self.easyvvuq_version
        }

        return out_dict



import os
import json
from easyvvuq.encoders import BaseEncoder

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class GaussEncoder(BaseEncoder, encoder_name="gauss"):

    def __init__(self, target_filename="gauss_input.json"):
        self.target_filename = target_filename

    def encode(self, params={}, target_dir=''):

        out_file = params['out_file']
        num_steps = params['num_steps']
        mu = params['mu']
        sigma = params['sigma']

        output_str = (f'{{"outfile": "{out_file}", "num_steps": "{num_steps}",'
                      f' "mu": "{mu}", "sigma": "{sigma}"}}\n')

        target_file_path = os.path.join(target_dir, self.target_filename)
        with open(target_file_path, "w") as outfile:
            outfile.write(output_str)

    def get_restart_dict(self):
        return {"target_filename": self.target_filename}

    def element_version(self):
        return "0.1"



from easyvvuq.decoders.json import JSONDecoder
import os
import numpy as np
import pytest


def test_jsondecoder_basic():
    decoder = JSONDecoder(os.path.join('jsondecoder', 'fredrik.json'), ['cfrac', 'we', 'v'])
    run_info = {'run_dir': 'tests'}
    data = decoder.parse_sim_output(run_info)
    assert(data['cfrac'] == 0.24000000131541285)
    assert(data['we'] == -0.4910355508327484)
    assert(len(data['v']) == 126)
    assert(data['v'][:3] == [0.014841768890619278, 0.014779693447053432, 0.014733896590769291])
    assert(data['v'][-3:] == [0.0010381652973592281, 0.0010054642334580421, 0.0009733123588375747])


def test_jsondecoder_scalars_only():
    # like test_jsondecoder_basics, but with only scalar quantities
    decoder = JSONDecoder(os.path.join('jsondecoder', 'fredrik.json'), ['cfrac', 'we'])
    run_info = {'run_dir': 'tests'}
    data = decoder.parse_sim_output(run_info)
    assert(data['cfrac'] == 0.24000000131541285)
    assert(data['we'] == -0.4910355508327484)


def test_json_nested():
    decoder = JSONDecoder(os.path.join('jsondecoder', 'nested.json'),
                          [['root1', 'node1', 'leaf1'], ['root1', 'leaf2'], 'leaf3'])
    run_info = {'run_dir': 'tests'}
    data = decoder.parse_sim_output(run_info)
    assert(data['root1.node1.leaf1'] == 0.33)
    assert(data['root1.leaf2'] == 0.32)
    assert(data['leaf3'] == [0.2, 0.3])


def test_missing_column():
    # Check if a RuntimeError is raised if a wrong column is specified
    decoder = JSONDecoder(os.path.join('jsondecoder', 'nested.json'),
                          [['root1', 'node1', 'abcd'], ['root1', 'leaf2'], 'leaf3'])
    run_info = {'run_dir': 'tests'}
    with pytest.raises(RuntimeError) as excinfo:
        data = decoder.parse_sim_output(run_info)
    # Check if the missing column is reported in the exception message
    assert("['root1', 'node1', 'abcd']" in str(excinfo.value))


def test_sim_complete():
    decoder = JSONDecoder('nested.json',
                          [['root1', 'node1', 'leaf1'], ['root1', 'leaf2'], 'leaf3'])
    assert(decoder.sim_complete({'run_dir': os.path.join('tests', 'jsondecoder')}))


def test_init_exceptions():
    with pytest.raises(RuntimeError):
        JSONDecoder(None, output_columns=[['root1', 'node1', 'leaf1'],
                                          ['root1', 'leaf2'], 'leaf3'])
    with pytest.raises(RuntimeError):
        JSONDecoder('nested.json', None)
    with pytest.raises(RuntimeError):
        JSONDecoder('nested.json', [])



from .base import BaseEncoder
import logging

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class MultiEncoder(BaseEncoder, encoder_name="multiencoder"):

    def __init__(self, *encoders):
        """
            Expects one or more encoders
        """
        self.encoders = encoders

    def encode(self, params={}, target_dir=''):
        """
            Applies all encoders in the list of encoders.
        """
        for encoder in self.encoders:
            encoder.encode(params=params, target_dir=target_dir)

    def is_restartable(self):
        return True



"""Provides element to execute a shell command in a given directory.
"""

import os
from pathlib import Path
import shutil
import subprocess
import dill
import copy

__license__ = "LGPL"


def local_execute(encoder, command, decoder, root='/tmp'):
    """A helper function for a simple local execution.
    It will create a directory under your specified root folder, encode the sampler output, execute a command
    and decode the results of the simulation.

    Parameters
    ----------
    encoder: Encoder
      an encoder to use
    command: list of str
      a command to run your simulation (same as argument to popen, e.g. ['ls', '-al'])
    decoder: Decoder
      a decoder to use
    root: str
      root folder, for example '/tmp' or if you want to use ram based filesystem it could be '/dev/shm'
  
    Returns
    -------
    EasyVVUQ Actions
    """
    return Actions(CreateRunDirectory(root), Encode(encoder), ExecuteLocal(command), Decode(decoder))


class CreateRunDirectory():
    def __init__(self, root):
        self.root = root

    def start(self, previous=None):
        run_id = previous['run_id']
        level1_a, level1_b = (int(run_id / 100 ** 4) * 100 ** 4,
                              int(run_id / 100 ** 4 + 1) * 100 ** 4)
        level2_a, level2_b = (int(run_id / 100 ** 3) * 100 ** 3,
                              int(run_id / 100 ** 3 + 1) * 100 ** 3)
        level3_a, level3_b = (int(run_id / 100 ** 2) * 100 ** 2,
                              int(run_id / 100 ** 2 + 1) * 100 ** 2)
        level4_a, level4_b = (int(run_id / 100 ** 1) * 100 ** 1,
                              int(run_id / 100 ** 1 + 1) * 100 ** 1)
        level1_dir = "runs_{}-{}/".format(level1_a, level1_b)
        level2_dir = "runs_{}-{}/".format(level2_a, level2_b)
        level3_dir = "runs_{}-{}/".format(level3_a, level3_b)
        level4_dir = "runs_{}-{}/".format(level4_a, level4_b)
        level5_dir = "runs_{}".format(int(run_id))
        path = os.path.join(self.root, previous['campaign_dir'], 'runs',
                            level1_dir, level2_dir, level3_dir, level4_dir, level5_dir)
        Path(path).mkdir(parents=True, exist_ok=True)
        previous['rundir'] = path
        self.result = previous
        return self.result

    def finished(self):
        return True

    def finalise(self):
        pass

    def succeeded(self):
        return True

class Encode():
    def __init__(self, encoder):
        self.encoder = encoder

    def start(self, previous=None):
        self.encoder.encode(
            params=previous['run_info']['params'],
            target_dir=previous['rundir'])
        try:
            previous['encoder_filename'] = self.encoder.target_filename
        except AttributeError:
            pass
        return previous

    def finished(self):
        return True

    def finalise(self):
        pass

    def succeeeded(self):
        return True

class Decode():
    def __init__(self, decoder):
        self.decoder = decoder

    def start(self, previous=None):
        run_info = copy.copy(previous['run_info'])
        run_info['run_dir'] = previous['rundir']
        result = self.decoder.parse_sim_output(run_info)
        previous['result'] = result
        previous['decoder_filename'] = self.decoder.target_filename
        return previous

    def finished(self):
        return True

    def finalise(self):
        pass

    def succeeded(self):
        return True

class CleanUp():
    def __init__(self):
        pass

    def start(self, previous=None):
        if not ('rundir' in previous.keys()):
            raise RuntimeError('must be used with actions that create a directory structure')
        shutil.rmtree(previous['rundir'])
        return previous
                
    def finished(self):
        return True

    def finalise(self):
        pass

    def succeeded(self):
        return True

class ExecutePython():
    def __init__(self, function):
        self.function = dill.dumps(function)
        self.params = None
        self.eval_result = None

    def start(self, previous=None):
        function = dill.loads(self.function)
        self.eval_result = function(previous['run_info']['params'])
        previous['result'] = self.eval_result
        return previous

    def finished(self):
        if self.eval_result is None:
            return False
        else:
            return True

    def finalise(self):
        pass

    def succeeded(self):
        if not self.finished():
            raise RuntimeError('action did not finish yet')
        else:
            return True

class ExecuteLocal():
    def __init__(self, full_cmd):
        self.full_cmd = full_cmd.split()
        self.popen_object = None
        self.ret = None
        self._started = False

    def start(self, previous=None):
        target_dir = previous['rundir']
        self.ret = subprocess.run(self.full_cmd, cwd=target_dir)
        return previous

    def finished(self):
        return True

    def finalise(self):
        """Performs clean-up if necessary. In this case it isn't. I think.
        """
        pass

    def succeeded(self):
        """Will return True if the process finished successfully.
        It judges based on the return code and will return False
        if that code is not zero.
        """
        if self.ret != 0:
            return False
        else:
            return True

class Actions():
    def __init__(self, *args):
        self.actions = list(args)

    def start(self, previous=None):
        for action in self.actions:
            if not hasattr(action, 'start'):
                raise RuntimeError('action in the actions list does not provide a start method')
        previous = copy.copy(previous)
        run_id = previous['run_id']
        for action in self.actions:
            previous = action.start(previous)
        self.result = previous
        assert(self.result['run_id'] == run_id)
        return previous

    def finished(self):
        return all([action.finished() for action in self.actions])

    def finalise(self):
        for action in self.actions:
            action.finalise()

    def succeeded(self):
        return all([action.succeeded() for action in self.actions])



import chaospy as cp
import easyvvuq as uq
import pytest
import yaml
import pickle
import numpy as np

# No longer required
# def test_l_n_exception():
#     vary = {
#         "gravity": cp.Uniform(9.8, 1.0),
#         "mass": cp.Uniform(2.0, 10.0),
#     }
#     with pytest.raises(RuntimeError):
#         sampler = uq.sampling.SCSampler(vary=vary, polynomial_order=1, sparse=True)


def test_lagrange_poly():
    assert(uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 0) == -3.5)
    assert(uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 1) == 2.0999999999999996)
    assert(uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 2) == 2.4000000000000004)
    with pytest.raises(IndexError):
        uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 3)



"""Provides analysis element for ensemble bootstrapping analysis.
"""
import numpy as np
import pandas as pd
from easyvvuq import OutputType
from .base import BaseAnalysisElement

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


def confidence_interval(dist, value, alpha, pivotal=False):
    """
    Get the bootstrap confidence interval for a given distribution.

    Parameters
    ----------
    dist:
        Array containing distribution of bootstrap results.
    value:
        Value of statistic for which we are calculating error bars.
    alpha:
        The alpha value for the confidence intervals.
    pivotal:
        Use the pivotal method? Default to percentile method.

    Returns
    -------

    float:
          Value of the bootstrap statistic
    float:
          Highest value of the confidence interval
    float:
          Lowest value of the confidence interval

    """
    if len(dist) < 1:
        raise ValueError("Dist array should be non-empty")

    if pivotal:

        low = 2 * value - np.percentile(dist, 100 * (1 - alpha / 2.), axis=0)
        stat = value
        high = 2 * value - np.percentile(dist, 100 * (alpha / 2.), axis=0)

    else:

        low = np.percentile(dist, 100 * (alpha / 2.), axis=0)
        stat = np.percentile(dist, 50)
        high = np.percentile(dist, 100 * (1 - alpha / 2.), axis=0)

    # if low > high:
    #     (low, high) = (high, low)

    return stat, low, high


def bootstrap(data, stat_func, alpha=0.05,
              sample_size=None, n_samples=1000,
              pivotal=False):
    """

    Parameters
    ----------
    data : :obj:`pandas.DataFrame`
        Input data to be analysed.
    stat_func : function
        Statistical function to be applied to data for bootstrapping.
    alpha : float
        Produce estimate of 100.0*(1-`alpha`) confidence interval.
    sample_size : int
        Size of the sample to be drawn from the input data.
    n_samples : int
        Number of times samples are to be drawn from the input data.
    pivotal : bool
        Use the pivotal method? Default to percentile method.

    Returns
    -------
    float:
          Value of the bootstrap statistic
    float:
          Highest value of the confidence interval
    float:
          Lowest value of the confidence interval
    """
    if data.empty:
        raise RuntimeError("DataFrame passed to bootstrap has to be non-empty")

    stat = data.apply(stat_func)

    if sample_size is None:
        sample_size = len(data)

    dist = []

    for l in range(n_samples):

        sample = data.sample(sample_size, replace=True)

        dist.append(stat_func(sample))

    return confidence_interval(dist, stat, alpha, pivotal=pivotal)


def ensemble_bootstrap(data, groupby=[], qoi_cols=[],
                       stat_func=np.mean, alpha=0.05,
                       sample_size=None, n_samples=1000,
                       pivotal=False, stat_name='boot'):
    """
    Perform bootstrapping analysis on input data.

    Parameters
    ----------
    data : :obj:`pandas.DataFrame`
        DataFrame to be analysed.
    groupby : list or None
        Columns to use to group the data in `analyse` method before
        calculating stats.
    qoi_cols : list or None
        Columns of quantities of interest (for which stats will be
        calculated).
    stat_func : function
        Statistical function to be applied to data for bootstrapping.
    alpha : float, default=0.05
        Produce estimate of 100.0*(1-`alpha`) confidence interval.
    sample_size : int
        Size of the sample to be drawn from the input data.
    n_samples : int, default=1000
        Number of times samples are to be drawn from the input data.
    pivotal : bool, default=False
        Use the pivotal method? Default to percentile method.
    stat_name : str, default='boot'
        Name to use to describe columns containing output statistic (for example
        'mean').

    Returns
    -------
    :obj:`pandas.DataFrame`
        Description of input data using bootstrap statistic and high/low
        confidence intervals.
    """

    agg_funcs = {}

    if not qoi_cols:
        qoi_cols = [
            x for x in data.columns if x not in groupby + ['run_id', 'status']]

    for col in qoi_cols:
        if col not in data:
            raise RuntimeError(f"No such attribute: {col}\nAttributes found in data: {data}")
        agg_funcs[col] = lambda x: bootstrap(
            x,
            stat_func=stat_func,
            alpha=alpha,
            sample_size=sample_size,
            n_samples=n_samples,
            pivotal=pivotal)

    if not groupby:
        grouped_data = data.groupby(lambda x: True, sort=False)
    else:
        grouped_data = data.groupby(groupby, sort=False)

    # Apply bootstrapping to all value columns selected
    # Note results come a tuple per cell
    results = grouped_data.agg(agg_funcs)

    outputs = [stat_name, 'low', 'high']

    # Split out tuples in each cell and provide sensible naming
    results = pd.concat({col: results[col].apply(
        lambda cell: pd.Series(cell, index=outputs)
    )
        for col in qoi_cols}, axis=1)

    return results


class EnsembleBoot(BaseAnalysisElement):

    def __init__(self, groupby=[], qoi_cols=[],
                 stat_func=np.mean, alpha=0.05,
                 sample_size=None, n_boot_samples=1000,
                 pivotal=False, stat_name='boot'):
        """
        Element to perform bootstrapping on collated simulation output.

        Parameters
        ----------
        groupby : list or None
            Columns to use to group the data in `analyse` method before
            calculating stats.
        qoi_cols : list or None
            Columns of quantities of interest (for which stats will be
            calculated).
        stat_func : function
            Statistical function to be applied to data for bootstrapping.
        alpha : float, default=0.05
            Produce estimate of 100.0*(1-`alpha`) confidence interval.
        sample_size : int
            Size of the sample to be drawn from the input data.
        n_boot_samples : int, default=1000
            Number of times samples are to be drawn from the input data.
        pivotal : bool, default=False
            Use the pivotal method? Default to percentile method.
        stat_name : str, default='boot'
            Name to use to describe columns containing output statistic (for example
            'mean').
        """

        self.groupby = groupby
        self.qoi_cols = qoi_cols

        self.stat_func = stat_func
        self.alpha = alpha
        self.sample_size = sample_size
        self.n_boot_samples = n_boot_samples
        self.pivotal = pivotal
        self.stat_name = stat_name

        self.output_type = OutputType.SUMMARY

        if self.stat_func is None:
            raise ValueError('stat_func cannot be None.')

    def element_name(self):
        """Name for this element for logging purposes"""
        return "ensemble_boot"

    def element_version(self):
        """Version of this element for logging purposes"""
        return "0.1"

    def analyse(self, data_frame=None):
        """Perform bootstrapping analysis on the input `data_frame`.

        The data_frame is grouped according to `self.groupby` if specified and
        analysis is performed on the columns selected in `self.qoi_cols` if set.

        Parameters
        ----------
        data_frame : :obj:`pandas.DataFrame`
            Summary data produced through collation of simulation output.

        Returns
        -------
        :obj:`pandas.DataFrame`
            Basic statistic for selected columns and groupings of data.
        """

        if data_frame is None:
            raise RuntimeError(
                "This VVUQ element needs a data frame to analyse")
        elif data_frame.empty:
            raise RuntimeError(
                "No data in data frame passed to analyse element")

        results = ensemble_bootstrap(
            data_frame,
            groupby=self.groupby,
            qoi_cols=self.qoi_cols,
            stat_func=self.stat_func,
            alpha=self.alpha,
            sample_size=self.sample_size,
            n_samples=self.n_boot_samples,
            pivotal=self.pivotal,
            stat_name=self.stat_name)

        return results



from ast import literal_eval
import dill
import base64


def easyvvuq_serialize(obj):
    """Takes an object and returns a string with that object serialized.

    Parameters
    ----------
    obj: obj
        An arbitrary Python object.

    Returns
    -------
    str:
        A string representation of obj.
    """
    return base64.b64encode(dill.dumps(obj)).decode('utf-8')


def easyvvuq_deserialize(s):
    """Takes a serialized objected and reconstructs it.

    Parameters
    ----------
    s: str
        A serialized Python object.

    Returns
    -------
    obj:
        A previously serialized Python object.
    """
    if isinstance(s, bytes):
        return dill.loads(base64.b64decode(s))
    else:
        return dill.loads(base64.b64decode(s.encode('utf-8')))


def multi_index_tuple_parser(lst):
    """
    Parses a list of strings to tuples if they represent tuples, otherwise
    leaves them as is.

    Parameters
    __________
    lst : list of strings

    Returns
    _______
    a tuple consisting of a list of tuples and/or strings and a bool indicating if
    the lst contains any tuples
    """
    contains_tuples = True
    if len(lst) == 0:
        raise RuntimeError('multi_index_tuple_parser needs a non-empty list of strings')
    for name in lst:
        if not isinstance(name, str):
            raise RuntimeError('multi_index_tuple_parser needs a list of strings')
    result = []
    for name in lst:
        if name.strip()[0] == '(' and name.strip()[-1] == ')':
            name = literal_eval(name)
        else:
            contains_tuples = False
        result.append(name)
    return result, contains_tuples


def remove_start_of_file(filename, delimiter):
    """Overwrite a file leaving only the portion of it after the
    delimiter string.

    Parameters
    ----------
    filename : str
       the name (and path) of the file to processs

    delimiter : str
       the string after which the file is copied over
    """
    with open(filename, 'r') as fd:
        lines = fd.readlines()
    remaining = []
    after_delimiter = False
    for line in lines:
        if after_delimiter:
            remaining.append(line)
        if line.strip() == delimiter:
            after_delimiter = True
    with open(filename, 'w') as fd:
        for line in remaining:
            fd.write(line)



"""Provides a base class for all analysis elements.
"""
from .. import BaseElement

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class BaseAnalysisElement(BaseElement):
    """Base class for all EasyVVUQ analysis elements.

    Attributes
    ----------

    """

    def analyse(self, data_frame=None):
        """Perform analysis on input `data_frame`.

        Parameters
        ----------
        data_frame : pandas DataFrame
            Input data for analysis.

        Returns
        -------
        AnalysisResults instance
        """
        raise NotImplementedError

    def element_category(self):
        """Element type for logging and verification.

        Returns
        -------
        str
            Element category.
        """
        return "analysis"

    def element_name(self):
        """Name for this element for logging purposes.

        Returns
        -------
        str
            Element name.
        """
        raise NotImplementedError

    def element_version(self):
        """Version of this element for logging purposes.

        Returns
        -------
        str
            Element version.
        """
        raise NotImplementedError



import pytest
import time
import chaospy as cp
from unittest.mock import MagicMock
from easyvvuq.actions import ActionPool, ExecutePython, Actions
from easyvvuq import Campaign
from easyvvuq.sampling import RandomSampler

@pytest.fixture
def campaign():
    def model(params):
        return {'y' : params['x'] + 1}
    actions = Actions(ExecutePython(model))
    sampler = RandomSampler({'x': cp.Uniform(0, 1)})
    campaign = Campaign('test', {'x': {'default' : 0}}, actions)
    campaign.set_sampler(sampler)
    return campaign

def test_action_pool_start(campaign):
    action_pool = campaign.execute(nsamples=3)
    assert(len(action_pool.futures) == 3)
    action_pool.collate()
    assert(len(action_pool.campaign.get_collation_result()) == 3)
    assert(action_pool.progress() == {'ready': 0, 'active': 0, 'finished': 3, 'failed': 0})

def test_action_pool_start_sequential(campaign):
    action_pool = campaign.execute(nsamples=3, sequential=True)
    assert(len(action_pool.results) == 3)
    action_pool.collate()
    assert(len(action_pool.campaign.get_collation_result()) == 3)



# for the Sobol g function, the exact (1st-order)
# Sobol indices are known analytically
import os
import easyvvuq as uq
import numpy as np
import chaospy as cp


def print_exact_sobols():
    V_i = np.zeros(d)

    for i in range(d):
        V_i[i] = 1.0 / (3.0 * (1 + a[i])**2)

    V = np.prod(1 + V_i) - 1

    print('----------------------')
    print('Exact 1st-order Sobol indices: ', V_i / V)


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    plt.close('all')

    # number of unknown variables
    d = 5

    # parameters required by test function
    a = [0.0, 0.5, 3.0, 9.0, 99.0]

    # author: Wouter Edeling
    __license__ = "LGPL"

    # home directory of user
    home = os.path.expanduser('~')
    HOME = os.path.abspath(os.path.dirname(__file__))

    # Set up a fresh campaign called "sc"
    my_campaign = uq.Campaign(name='sc', work_dir='/tmp')

    # Define parameter space
    params = {
        "x1": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.5},
        "x2": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.5},
        "x3": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.5},
        "x4": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.5},
        "x5": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.5},
        "out_file": {
            "type": "string",
            "default": "output.csv"}}

    output_filename = params["out_file"]["default"]
    output_columns = ["f"]

    # Create an encoder, decoder and collation element
    encoder = uq.encoders.GenericEncoder(
        template_fname=HOME + '/sobol.template',
        delimiter='$',
        target_filename='sobol_in.json')
    decoder = uq.decoders.SimpleCSV(target_filename=output_filename,
                                    output_columns=output_columns)

    # Add the SC app (automatically set as current app)
    my_campaign.add_app(name="sc",
                        params=params,
                        encoder=encoder,
                        decoder=decoder)

    # Create the sampler
    vary = {
        "x1": cp.Uniform(0.0, 1.0),
        "x2": cp.Uniform(0.0, 1.0),
        "x3": cp.Uniform(0.0, 1.0),
        "x4": cp.Uniform(0.0, 1.0),
        "x5": cp.Uniform(0.0, 1.0)}

    """
    SPARSE GRID PARAMETERS
    ----------------------
    - sparse = True: use a Smolyak sparse grid
    - growth = True: use an exponential rule for the growth of the number
      of 1D collocation points per level. Used to make e.g. clenshaw-curtis
      quadrature nested.
    """
    my_sampler = uq.sampling.SCSampler(vary=vary, polynomial_order=2,
                                       quadrature_rule="G", sparse=False,
                                       growth=False)

    # Associate the sampler with the campaign
    my_campaign.set_sampler(my_sampler)

    # Will draw all (of the finite set of samples)
    my_campaign.draw_samples()
    my_campaign.populate_runs_dir()

    # Use this instead to run the samples using EasyVVUQ on the localhost
    my_campaign.apply_for_each_run_dir(uq.actions.ExecuteLocal(
        "tests/sc/sobol_model.py sobol_in.json"))

    my_campaign.collate()

    # Post-processing analysis
    analysis = uq.analysis.SCAnalysis(sampler=my_sampler, qoi_cols=output_columns)

    my_campaign.apply_analysis(analysis)

    results = my_campaign.get_last_analysis()

    print(results.sobols_first())

    print_exact_sobols()

    plt.show()



import pytest
import os
from easyvvuq.utils.helpers import multi_index_tuple_parser, remove_start_of_file


def test_multi_index_tuple_parser_exceptions():
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser([])
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser([3])
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser(['a', 'b', 3])
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser([3, 'a', 'b'])


def test_multi_index_parser_corner_cases():
    assert(multi_index_tuple_parser(["a"]) == (["a"], False))
    assert(multi_index_tuple_parser(['a', '("b", 1)']) == (["a", ("b", 1)], False))


def test_multi_index_parser():
    assert(multi_index_tuple_parser(['("a", 1)', '("b", 1)']) == ([("a", 1), ("b", 1)], True))


def test_remove_start_of_file(tmp_path):
    data = """alkjd
    dklfjf
     d dsf dsfjdsl
    START
    sdlf fsd
    fdsj
    """
    trimmed = """    sdlf fsd
    fdsj
    """
    with open(os.path.join(tmp_path, 'test.txt'), 'w') as fd:
        fd.write(data)
    remove_start_of_file(os.path.join(tmp_path, 'test.txt'), 'START')
    with open(os.path.join(tmp_path, 'test.txt'), 'r') as fd:
        assert(fd.read() == trimmed)



import pytest
import time
from unittest.mock import MagicMock
from easyvvuq.actions import ActionPool

@pytest.fixture(scope="module", params=[False])
def action_pool():
    campaign = MagicMock()
    actions = MagicMock()
    inits = [MagicMock(), MagicMock(), MagicMock()]
    return ActionPool(campaign, actions, inits, sequential=False)

def test_action_pool_start(action_pool):
    pass



"""Analysis element for polynomial chaos expansion (PCE). We use ChaosPy
under the hood for this functionality.
"""
import logging
import chaospy as cp
import numpy as np
from easyvvuq import OutputType
from .base import BaseAnalysisElement
from .results import AnalysisResults
from .qmc_analysis import QMCAnalysisResults

__author__ = 'Jalal Lakhlili'
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class PCEAnalysisResults(QMCAnalysisResults):
    """Analysis results for the PCEAnalysis class.
    """

    def _get_sobols_first(self, qoi, input_):
        """Returns the first order sobol index for a given qoi wrt input variable.

        Parameters
        ----------
        qoi : str
           Quantity of interest
        input_ : str
           Input variable

        Returns
        -------
        float
            First order sobol index.
        """
        raw_dict = AnalysisResults._keys_to_tuples(self.raw_data['sobols_first'])
        return raw_dict[AnalysisResults._to_tuple(qoi)][input_]

    def _get_sobols_second(self, qoi, input_):
        """Returns the second order sobol index for a given qoi wrt input variable.

        Parameters
        ----------
        qoi : str
           Quantity of interest
        input_ : str
           Input variable

        Returns
        -------
        float
            Second order sobol index.
        """
        raw_dict = AnalysisResults._keys_to_tuples(self.raw_data['sobols_second'])
        return dict([(in_, raw_dict[AnalysisResults._to_tuple(qoi)][input_][i])
                     for i, in_ in enumerate(self.inputs) if in_ != input_])

    def _get_sobols_total(self, qoi, input_):
        """Returns the total order sobol index for a given qoi wrt input variable.

        Parameters
        ----------
        qoi : str
           Quantity of interest
        input_ : str
           Input variable

        Returns
        -------
        float
            Total order sobol index.
        """
        raw_dict = AnalysisResults._keys_to_tuples(self.raw_data['sobols_total'])
        return raw_dict[AnalysisResults._to_tuple(qoi)][input_]

    def _describe(self, qoi, statistic):
        """Returns descriptive statistics, similar to pandas describe.

        Parameters
        ----------
        qoi : str
            Name of quantity of interest.
        statistic : str
            One of 'min', 'max', '10%', '90%', 'median', 'mean', 'var', 'std'

        Returns
        -------
        float
            Value of the requested statistic.
        """
        if statistic == 'min':
            return np.array([v.lower[0] for _, v in enumerate(
                self.raw_data['output_distributions'][qoi])])
        elif statistic == 'max':
            return np.array([v.upper[0] for _, v in enumerate(
                self.raw_data['output_distributions'][qoi])])
        elif statistic == '1%':
            return self.raw_data['percentiles'][qoi]['p01']
        elif statistic == '10%':
            return self.raw_data['percentiles'][qoi]['p10']
        elif statistic == '90%':
            return self.raw_data['percentiles'][qoi]['p90']
        elif statistic == '99%':
            return self.raw_data['percentiles'][qoi]['p99']
        elif statistic == 'median':
            return self.raw_data['percentiles'][qoi]['p50']
        else:
            try:
                return self.raw_data['statistical_moments'][qoi][statistic]
            except KeyError:
                raise NotImplementedError


class PCEAnalysis(BaseAnalysisElement):

    def __init__(self, sampler=None, qoi_cols=None, sampling=False):
        """Analysis element for polynomial chaos expansion (PCE).

        Parameters
        ----------
        sampler : PCESampler
            Sampler used to initiate the PCE analysis.
        qoi_cols : list or None
            Column names for quantities of interest (for which analysis is
            performed).
        sampling : True if chaospy sampling method to be used for calculating
            statitical quantities; otherwise [default] the pce coefficients are used
        """

        if sampler is None:
            msg = 'PCE analysis requires a paired sampler to be passed'
            raise RuntimeError(msg)

        if qoi_cols is None:
            raise RuntimeError("Analysis element requires a list of "
                               "quantities of interest (qoi)")

        self.qoi_cols = qoi_cols
        self.sampling = sampling
        self.output_type = OutputType.SUMMARY
        self.sampler = sampler

    def element_name(self):
        """Name for this element for logging purposes.

        Returns
        -------
        str
            "PCE_Analysis"
        """
        return "PCE_Analysis"

    def element_version(self):
        """Version of this element for logging purposes.

        Returns
        -------
        str
            Element version.
        """
        return "0.6"

    def analyse(self, data_frame=None):
        """Perform PCE analysis on input `data_frame`.

        Parameters
        ----------
        data_frame : pandas DataFrame
            Input data for analysis.

        Returns
        -------
        PCEAnalysisResults
            Use it to get the sobol indices and other information.
        """

        def sobols(P, coefficients):
            """ Utility routine to calculate sobols based on coefficients
            """
            A = np.array(P.coefficients) != 0
            multi_indices = np.array([P.exponents[A[:, i]].sum(axis=0) for i in range(A.shape[1])])
            sobol_mask = multi_indices != 0
            _, index = np.unique(sobol_mask, axis=0, return_index=True)
            index = np.sort(index)
            sobol_idx_bool = sobol_mask[index]
            sobol_idx_bool = np.delete(sobol_idx_bool, [0], axis=0)
            n_sobol_available = sobol_idx_bool.shape[0]
            if len(coefficients.shape) == 1:
                n_out = 1
            else:
                n_out = coefficients.shape[1]
            n_coeffs = coefficients.shape[0]
            sobol_poly_idx = np.zeros([n_coeffs, n_sobol_available])
            for i_sobol in range(n_sobol_available):
                sobol_poly_idx[:, i_sobol] = np.all(sobol_mask == sobol_idx_bool[i_sobol], axis=1)
            sobol = np.zeros([n_sobol_available, n_out])
            for i_sobol in range(n_sobol_available):
                sobol[i_sobol] = np.sum(
                    np.square(coefficients[sobol_poly_idx[:, i_sobol] == 1]), axis=0)
            idx_sort_descend_1st = np.argsort(sobol[:, 0], axis=0)[::-1]
            sobol = sobol[idx_sort_descend_1st, :]
            sobol_idx_bool = sobol_idx_bool[idx_sort_descend_1st]
            sobol_idx = [0 for _ in range(sobol_idx_bool.shape[0])]
            for i_sobol in range(sobol_idx_bool.shape[0]):
                sobol_idx[i_sobol] = np.array(
                    [i for i, x in enumerate(sobol_idx_bool[i_sobol, :]) if x])
            var = ((coefficients[1:]**2).sum(axis=0))
            sobol = sobol / var
            return sobol, sobol_idx, sobol_idx_bool

        if data_frame is None:
            raise RuntimeError("Analysis element needs a data frame to "
                               "analyse")
        elif data_frame.empty:
            raise RuntimeError(
                "No data in data frame passed to analyse element")

        qoi_cols = self.qoi_cols

        results = {'statistical_moments': {},
                   'percentiles': {},
                   'sobols_first': {k: {} for k in qoi_cols},
                   'sobols_second': {k: {} for k in qoi_cols},
                   'sobols_total': {k: {} for k in qoi_cols},
                   'correlation_matrices': {},
                   'output_distributions': {},
                   'fit': {},
                   'Fourier_coefficients': {},
                   }

        # Get sampler informations
        P = self.sampler.P
        nodes = self.sampler._nodes
        weights = self.sampler._weights
        regression = self.sampler.regression

        # Extract output values for each quantity of interest from Dataframe
#        samples = {k: [] for k in qoi_cols}
#        for run_id in data_frame[('run_id', 0)].unique():
#            for k in qoi_cols:
#                data = data_frame.loc[data_frame[('run_id', 0)] == run_id][k]
#                samples[k].append(data.values.flatten())

        samples = {k: [] for k in qoi_cols}
        for k in qoi_cols:
            samples[k] = data_frame[k].values

        # Compute descriptive statistics for each quantity of interest
        for k in qoi_cols:
            # Approximation solver
            if regression:
                fit, fc = cp.fit_regression(P, nodes, samples[k], retall=1)
            else:
                fit, fc = cp.fit_quadrature(P, nodes, weights, samples[k], retall=1)
            results['fit'][k] = fit
            results['Fourier_coefficients'][k] = fc

            # Percentiles: 1%, 10%, 50%, 90% and 99%
            P01, P10, P50, P90, P99 = cp.Perc(
                fit, [1, 10, 50, 90, 99], self.sampler.distribution).squeeze()
            results['percentiles'][k] = {'p01': P01, 'p10': P10, 'p50': P50, 'p90': P90, 'p99': P99}

            if self.sampling:  # use chaospy's sampling method

                # Statistical moments
                mean = cp.E(fit, self.sampler.distribution)
                var = cp.Var(fit, self.sampler.distribution)
                std = cp.Std(fit, self.sampler.distribution)
                results['statistical_moments'][k] = {'mean': mean,
                                                     'var': var,
                                                     'std': std}

                # Sensitivity Analysis: First, Second and Total Sobol indices
                sobols_first_narr = cp.Sens_m(fit, self.sampler.distribution)
                sobols_second_narr = cp.Sens_m2(fit, self.sampler.distribution)
                sobols_total_narr = cp.Sens_t(fit, self.sampler.distribution)
                sobols_first_dict = {}
                sobols_second_dict = {}
                sobols_total_dict = {}
                for i, param_name in enumerate(self.sampler.vary.vary_dict):
                    sobols_first_dict[param_name] = sobols_first_narr[i]
                    sobols_second_dict[param_name] = sobols_second_narr[i]
                    sobols_total_dict[param_name] = sobols_total_narr[i]

                results['sobols_first'][k] = sobols_first_dict
                results['sobols_second'][k] = sobols_second_dict
                results['sobols_total'][k] = sobols_total_dict

            else:  # use PCE coefficients

                # Statistical moments
                mean = fc[0]
                var = np.sum(fc[1:]**2, axis=0)
                std = np.sqrt(var)
                results['statistical_moments'][k] = {'mean': mean,
                                                     'var': var,
                                                     'std': std}

                # Sensitivity Analysis: First, Second and Total Sobol indices
                sobol, sobol_idx, _ = sobols(P, fc)
                varied = [_ for _ in self.sampler.vary.get_keys()]
                S1 = {_: np.zeros(sobol.shape[-1]) for _ in varied}
                ST = {_: np.zeros(sobol.shape[-1]) for _ in varied}
                #S2 = {_ : {__: np.zeros(sobol.shape[-1]) for __ in varied} for _ in varied}
                #for v in varied: del S2[v][v]
                S2 = {_: np.zeros((len(varied), sobol.shape[-1])) for _ in varied}
                for n, si in enumerate(sobol_idx):
                    if len(si) == 1:
                        v = varied[si[0]]
                        S1[v] = sobol[n]
                    elif len(si) == 2:
                        v1 = varied[si[0]]
                        v2 = varied[si[1]]
                        #S2[v1][v2] = sobol[n]
                        #S2[v2][v1] = sobol[n]
                        S2[v1][si[1]] = sobol[n]
                        S2[v2][si[0]] = sobol[n]
                    for i in si:
                        ST[varied[i]] += sobol[n]

                results['sobols_first'][k] = S1
                results['sobols_second'][k] = S2
                results['sobols_total'][k] = ST

            # Correlation matrix
            results['correlation_matrices'][k] = cp.Corr(
                fit, self.sampler.distribution)

            # Output distributions
            results['output_distributions'][k] = cp.QoI_Dist(
                fit, self.sampler.distribution)

        return PCEAnalysisResults(raw_data=results, samples=data_frame,
                                  qois=self.qoi_cols, inputs=list(self.sampler.vary.get_keys()))



"""Replica Sampler

Summary
-------

Primarily intended for sampling the same paramater values but with
different random seed. Other uses may be possible. It takes a finite
sampler and produces an infinite sampler from it. This infinite
sampler loops through the parameters produced by the finite sampler
and at each cycle adds a unique id number for that cycle to the
parameter dictionary.
"""

from easyvvuq.sampling import BaseSamplingElement
from itertools import cycle


class ReplicaSampler(BaseSamplingElement, sampler_name='replica_sampler'):
    """Replica Sampler

    Parameters
    ----------
    sampler : an instance of a class derived from  BaseSamplingElement
        a finite sampler to loop over

    replica_col : string
        a parameter name for the replica id
    seed_col : string
        a parameter name for the input parameter that specifies the RNG seed
    replicas : int
        number of replicas, if zero will result in an infinite sampler
    """

    def __init__(self, sampler, replica_col='ensemble_id', seed_col=None, replicas=0):
        if not sampler.is_finite():
            raise RuntimeError("Replica sampler only works with finite samplers")
        self.sampler = sampler
        self.replica_col = replica_col
        self.replicas = replicas
        self.sampler.n_replicas = replicas
        self.reset()

    def reset(self):
        self.history = []
        for sample in self.sampler:
            self.history.append(sample)
        self.size = len(self.history)
        self.cycle = cycle(self.history)
        self.counter = 0
        self.total_counter = self.replicas * self.sampler.n_samples()

    def is_finite(self):
        if self.replicas == 0:
            return False
        else:
            return True

    def n_samples(self):
        if self.replicas == 0:
            raise RuntimeError("You can't get the number of samples in an infinite sampler")
        else:
            return self.replicas * self.sampler.n_samples()

    def __next__(self):
        if self.replicas != 0:
            self.total_counter -= 1
            if self.total_counter < 0:
                raise StopIteration
        params = dict(next(self.cycle))
        params[self.replica_col] = self.counter
        self.counter = (self.counter + 1) % self.size
        return params

    def update(self, result, invalid):
        self.reset()
        return self.sampler.update(result, invalid)

    @property
    def iteration(self):
        return self.sampler.iteration

    @property
    def analysis_class(self):
        return self.sampler.analysis_class

    @property
    def inputs(self):
        return self.sampler.inputs

    @property
    def qoi(self):
        return self.sampler.qoi



from easyvvuq.decoders.yaml import YAMLDecoder
import os
import numpy as np
import pytest


@pytest.mark.parametrize("keys, vals", [
    (['a'], [0.0]),  # single scalar
    (['a', 'b', 'c'], [0.0, 1.0, 2.0]),  # multiple scalar
    (['d'], [[1.0, 2.0, 3.0]]),  # single vector
    (['d', 'e'], [[1.0, 2.0, 3.0], [0.0, 1.0, 2.0]]),  # multiple vector
    (['a', 'b', 'd', 'e'], [0.0, 1.0, [1.0, 2.0, 3.0], [0.0, 1.0, 2.0]]),
])
def test_yamldecoder_data(keys, vals):
    decoder = YAMLDecoder(os.path.join('yamldecoder', 'test.yml'), keys)
    run_info = {'run_dir': 'tests'}
    data = decoder.parse_sim_output(run_info)

    for k, v in zip(keys, vals):
        assert((data[k] == np.array([v])).all().all())


def test_yaml_nested():
    decoder = YAMLDecoder(os.path.join('yamldecoder', 'nested.yml'),
                          [['root1', 'node1', 'leaf1'], ['root1', 'leaf2'], 'leaf3'])
    run_info = {'run_dir': 'tests'}
    data = decoder.parse_sim_output(run_info)
    assert((data['root1.node1.leaf1'] == np.array([0.33])).all().all())
    assert((data['root1.leaf2'] == np.array([0.32])).all().all())
    assert((data['leaf3'] == np.array([0.2, 0.3])).all().all())


def test_sim_complete():
    decoder = YAMLDecoder('nested.yml',
                          [['root1', 'node1', 'leaf1'], ['root1', 'leaf2'], 'leaf3'])
    assert(decoder.sim_complete({'run_dir': os.path.join('tests', 'yamldecoder')}))


def test_init_exception():
    with pytest.raises(RuntimeError):
        YAMLDecoder(None, output_columns=[['root1', 'node1', 'leaf1'],
                                          ['root1', 'leaf2'], 'leaf3'])

    with pytest.raises(RuntimeError):
        YAMLDecoder('yamldecoder/nested.yaml', None)

    with pytest.raises(RuntimeError):
        YAMLDecoder('yamldecoder/nested.yaml', None)



import easyvvuq as uq
import chaospy as cp
import os
import sys
import pytest
import logging
from pprint import pformat, pprint
from .gauss.encoder_gauss import GaussEncoder
from .gauss.decoder_gauss import GaussDecoder
from easyvvuq.decoders.json import JSONDecoder
from easyvvuq.actions import Actions, Encode, Decode, CreateRunDirectory

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""


__license__ = "LGPL"


logging.basicConfig(level=logging.CRITICAL)


def test_gauss_vector_sc(tmpdir):
    # vector version of test_gauss
    # loads json output containing vector data from gauss test

    params = {
        "sigma": {
            "type": "float",
            "min": 0.0,
            "max": 100000.0,
            "default": 0.25
        },
        "mu": {
            "type": "float",
            "min": 0.0,
            "max": 100000.0,
            "default": 1
        },
        "num_steps": {
            "type": "integer",
            "min": 0,
            "max": 100000,
            "default": 10
        },
        "out_file": {
            "type": "string",
            "default": "output.csv"
        },
    }

    vary = {
        "mu": cp.Uniform(1.0, 100.0),
    }

    encoder = uq.encoders.GenericEncoder(template_fname='tests/gauss/gauss.template',
                                         target_filename='gauss_in.json')
    decoder = uq.decoders.SimpleCSV(target_filename="output.csv",
                                    output_columns=["numbers"])
    execute = uq.actions.ExecuteLocal(os.path.abspath("tests/gauss/gauss_json.py") + " gauss_in.json")
    actions = Actions(CreateRunDirectory('/tmp'), Encode(encoder), execute, Decode(decoder))
    sampler = uq.sampling.SCSampler(vary=vary, polynomial_order=4)
    my_campaign = uq.Campaign(name='gauss_vector', work_dir=tmpdir)
    my_campaign.add_app(name="gauss_vector",
                        params=params,
                        actions=actions)
    my_campaign.set_sampler(sampler)
    my_campaign.execute().collate()
    data = my_campaign.get_collation_result()
    analysis = uq.analysis.SCAnalysis(sampler=sampler, qoi_cols=["numbers"])
    my_campaign.apply_analysis(analysis)
    results = my_campaign.get_last_analysis()


def test_gauss_vector_pce(tmpdir):
    # vector version of test_gauss
    # loads json output containing vector data from gauss test

    params = {
        "sigma": {
            "type": "float",
            "min": 0.0,
            "max": 100000.0,
            "default": 0.25
        },
        "mu": {
            "type": "float",
            "min": 0.0,
            "max": 100000.0,
            "default": 1
        },
        "num_steps": {
            "type": "integer",
            "min": 0,
            "max": 100000,
            "default": 10
        },
        "out_file": {
            "type": "string",
            "default": "output.csv"
        },
    }

    vary = {
        "mu": cp.Uniform(1.0, 100.0),
    }

    encoder = uq.encoders.GenericEncoder(template_fname='tests/gauss/gauss.template',
                                         target_filename='gauss_in.json')
    #decoder = JSONDecoder(target_filename='output.csv.json', output_columns=['numbers'])
    decoder = uq.decoders.SimpleCSV(target_filename="output.csv",
                                    output_columns=["numbers"])
    execute = uq.actions.ExecuteLocal(os.path.abspath("tests/gauss/gauss_json.py") + " gauss_in.json")
    actions = Actions(CreateRunDirectory('/tmp'), Encode(encoder), execute, Decode(decoder))
    sampler = uq.sampling.PCESampler(vary=vary, polynomial_order=4)
    my_campaign = uq.Campaign(name='gauss_vector', work_dir=tmpdir)
    my_campaign.add_app(name="gauss_vector",
                        params=params,
                        actions=actions)
    my_campaign.set_sampler(sampler)
    my_campaign.execute().collate()

    data = my_campaign.get_collation_result()
    analysis = uq.analysis.PCEAnalysis(sampler=sampler, qoi_cols=["numbers"])
    my_campaign.apply_analysis(analysis)
    results = my_campaign.get_last_analysis()


if __name__ == "__main__":
    # test_gauss_vector_pce("/tmp")
    test_gauss_vector_sc("/tmp")



"""Constants and Enums to set defaults and constrain selections


Attributes
----------

default_campaign_prefix : str
    Text used to ensure campaign names are identifiable and somewhat human
    readable.
"""
from enum import Enum, IntEnum

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

default_campaign_prefix = 'EasyVVUQ_Campaign_'


class OutputType(Enum):
    """
    Types of data output by UQPs/VVPs
    """

    SAMPLE = 'sample'
    ARRAY = 'array'
    SAMPLE_ARRAY = 'sample_array'
    TRACK = 'track'
    SUMMARY = 'summary'


class Status(IntEnum):
    """
    Status of runs in the Run Table
    """

    NEW = 1
    ENCODED = 2
    COLLATED = 3
    IGNORED = 4
    INVALID = 5



from easyvvuq.base_element import BaseElement
import logging

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


# Dict to store all registered samplers (any class which extends
# BaseSamplingElement is automatically registered as a sampler)
AVAILABLE_SAMPLERS = {}


class BaseSamplingElement(BaseElement):
    """Baseclass for all EasyVVUQ sampling elements.

    Attributes
    ----------
    sampler_name : str
        Name of the particular sampler.

    """

    iteration = 0

    def __init_subclass__(cls, sampler_name, **kwargs):
        """
        Catch any new samplers (all samplers must inherit from
        BaseSamplingElement) and add them to the dict of available samplers.

        Parameters
        ----------
        sampler_name : str
            Name of the particular sampler.
        """

        super().__init_subclass__(**kwargs)

        cls.sampler_name = sampler_name

        # Register new sampler
        AVAILABLE_SAMPLERS[sampler_name] = cls

    def element_category(self):
        return "sampling"

    def element_name(self):
        return self.sampler_name

    def is_finite(self):
        raise NotImplementedError

    def n_samples(self):
        raise NotImplementedError

    def __iter__(self):
        """
        This method allows the sampler to be used as an iterator.
        The campaign object's draw_samples() method uses samplers
        in that manner.
        """
        return self

    def __next__(self):
        """
        This must be implemented by any sampler class.
        It should return the next run in the sequence.
        In the case of a finite sampler, when there are
        no more runs remaining, __next__() should
        raise a StopIteration exception
        """
        raise NotImplementedError

    @property
    def analysis_class(self):
        raise NotImplementedError

    @property
    def sampler_id(self):
        try:
            return self._sampler_id
        except AttributeError:
            raise RuntimeError('this sampler does not have an id assigned, run set_sampler first')

    @sampler_id.setter
    def sampler_id(self, val):
        self._sampler_id = val


class Vary:
    def __init__(self, vary_dict):
        if vary_dict is None:
            msg = ("'vary_dict' cannot be None. RandomSampler must be passed a "
                   "dict of the names of the parameters you want to vary, "
                   "and their corresponding distributions.")
            logging.error(msg)
            raise Exception(msg)
        if not isinstance(vary_dict, dict):
            msg = ("'vary_dict' must be a dictionary of the names of the "
                   "parameters you want to vary, and their corresponding "
                   "distributions.")
            logging.error(msg)
            raise Exception(msg)
        if len(vary_dict) == 0:
            msg = "'vary_dict' cannot be empty."
            logging.error(msg)
            raise Exception(msg)

        self.vary_dict = vary_dict

    def get_items(self):
        return self.vary_dict.items()

    def get_values(self):
        return self.vary_dict.values()

    def get_keys(self):
        return self.vary_dict.keys()

    def __str__(self):
        return self.vary_dict.__str__()



#!/usr/bin/env python3

import sys
import os
import numpy as np
import json

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


if len(sys.argv) != 2:
    sys.exit("Usage: python3 gauss_json.py JSONIN")

json_input = sys.argv[1]

if not os.path.isfile(json_input):
    sys.exit(json_input + " does not exist.")

with open(json_input, "r") as f:
    inputs = json.load(f)

mu = float(inputs['mu'])
sigma = float(inputs['sigma'])
num_steps = int(inputs['num_steps'])
output_filename = inputs['outfile']

if 'biasfile' in inputs:
    with open(inputs['biasfile']) as fin:
        line = fin.readline()

    bias = float(line.split()[0])
else:
    bias = 0


if num_steps <= 0:
    sys.exit("num_steps should be > 0")

numbers = np.random.normal(mu, sigma, num_steps)
numbers += bias
numbers_out = np.array(list(enumerate(numbers)))

#header = 'Step,Value'

#fmt = '%i,%f'
#np.savetxt(output_filename, numbers_out, fmt=fmt, header=header)

#json_output = {'numbers': list(numbers)}
# with open(output_filename + '.json', 'wt') as json_fp:
#    json.dump(json_output, json_fp)

# output csv file
np.savetxt(output_filename, numbers,
           delimiter=",", comments='',
           header='numbers')



import pytest
import easyvvuq as uq
import chaospy as cp
import numpy as np
from easyvvuq.actions import Actions, Encode, Decode, CreateRunDirectory


def pytest_namespace():
    return {'shared': None}


@pytest.mark.dependency()
def test_draw(benchmark):
    params = {
        "S0": {"type": "float", "default": 997},
        "I0": {"type": "float", "default": 3},
        "beta": {"type": "float", "default": 0.2},
        "gamma": {"type": "float", "default": 0.04, "min": 0.0, "max": 1.0},
        "iterations": {"type": "integer", "default": 100},
        "outfile": {"type": "string", "default": "output.csv"}
    }
    encoder = uq.encoders.GenericEncoder(
        template_fname='tutorials/sir.template',
        delimiter='$',
        target_filename='input.json')
    decoder = uq.decoders.SimpleCSV(target_filename='output.csv', output_columns=['I'])
    execute = uq.actions.ExecuteLocal('test')
    actions = Actions(execute)
    campaign = uq.Campaign(name='sir_benchmark', params=params, actions=actions)
    pytest.shared = campaign
    vary = {
        "beta": cp.Uniform(0.15, 0.25),
        "gamma": cp.Normal(0.04, 0.001),
    }
    campaign.set_sampler(uq.sampling.RandomSampler(vary=vary))
    benchmark(campaign.draw_samples, 10000)


def fake_results():
    counter = 1
    while True:
        yield ('Run_{}'.format(counter), {'values': list(np.random.random(size=100))})
        counter += 1


@pytest.mark.dependency(depends=['test_draw'])
def test_store_results(benchmark):
    iterator = fake_results()
    results = []
    for _ in range(10000):
        results.append(next(iterator))
    benchmark(pytest.shared.campaign_db.store_results, pytest.shared._active_app_name, results)


@pytest.mark.dependency(depends=['test_store_results'])
def test_get_collation_result(benchmark):
    benchmark(pytest.shared.get_collation_result)


@pytest.mark.dependency(depends=['test_get_collation_result'])
def test_draw_add(benchmark):
    benchmark(pytest.shared.draw_samples, 10000)


@pytest.mark.dependency(depends=['test_draw_add'])
def test_store_results_add(benchmark):
    iterator = fake_results()
    results = []
    for _ in range(10000):
        results.append(next(iterator))
    benchmark(pytest.shared.campaign_db.store_results, pytest.shared._active_app_name, results)


@pytest.mark.dependency(depends=['test_store_results_add'])
def test_get_collation_result_add(benchmark):
    benchmark(pytest.shared.get_collation_result)



"""This sampler is meant to be used with the QMC Analysis module.
"""

import chaospy as cp
from SALib.sample import saltelli
from .base import BaseSamplingElement, Vary
import logging

__author__ = "Jalal Lakhlili"
__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class QMCSampler(BaseSamplingElement, sampler_name="QMC_sampler"):
    def __init__(self, vary, n_mc_samples, count=0):
        """Create a Quasi Monte Carlo sampler.

        Parameters
        ----------

        vary: dict
            Expects a dictionary where the keys are variable names
            (inputs for your simulation that you want to vary during
            sampling) and values are ChaosPy distributions you want to
            sample from.
        n_mc_samples : int
            An estimate for how many samples the monte carlo run will need.
        count : int
            This is used to resume sampling. It will skip the first
            count samples if this parameter is not zero.
        """
        if not isinstance(vary, dict):
            msg = ("'vary' must be a dictionary of the names of the "
                   "parameters you want to vary, and their corresponding "
                   "distributions.")
            raise RuntimeError(msg)
        if len(vary) == 0:
            msg = "'vary' cannot be empty."
            raise RuntimeError(msg)

        self.vary = Vary(vary)
        self.n_mc_samples = n_mc_samples

        # List of the probability distributions of uncertain parameters
        params_distribution = list(vary.values())

        # Multivariate distribution
        self.distribution = cp.J(*params_distribution)

        # Generate samples
        self.n_params = len(vary)

        dist_U = []
        for i in range(self.n_params):
            dist_U.append(cp.Uniform())
        dist_U = cp.J(*dist_U)

        problem = {
            "num_vars": self.n_params,
            "names": list(vary.keys()),
            "bounds": [[0, 1]] * self.n_params
        }

        nodes = saltelli.sample(problem, n_mc_samples, calc_second_order=False)

        self._samples = self.distribution.inv(dist_U.fwd(nodes.transpose()))

        self._n_samples = n_mc_samples * (self.n_params + 2)

        # Fast forward to specified count, if possible
        self.count = 0
        if count >= self._n_samples:
            msg = (f"Attempt to start sampler fastforwarded to count {count}, "
                   f"but sampler only has {self._n_samples} samples, therefore"
                   f"this sampler will not provide any more samples.")
            logging.debug(msg)
        else:
            for _ in range(count):
                self.__next__()

    def is_finite(self):
        """Can this sampler produce only a finite number of samples."""
        return True

    @property
    def n_samples(self):
        """Returns the number of samples in this sampler.

        Returns
        -------
        This computed with the formula (d + 2) * N, where d is the number
        of uncertain parameters and N is the (estimated) number of samples
        for the Monte Carlo method.
        """
        return self._n_samples

    @property
    def analysis_class(self):
        """Return a corresponding analysis class.
        """
        from easyvvuq.analysis import QMCAnalysis
        return QMCAnalysis

    def __next__(self):
        if self.count < self.n_samples:
            run_dict = {}
            i_par = 0
            for param_name in self.vary.get_keys():
                run_dict[param_name] = self._samples.T[self.count][i_par]
                i_par += 1
            self.count += 1
            return run_dict
        else:
            raise StopIteration



"""A CSV file based sampler.

Useful for cases where you want to evaluate a sampling plan generated by other software.
Will take a CSV file with an appropriate header and will output it row by row.
"""

from .base import BaseSamplingElement
import csv

__license__ = "LGPL"


class CSVSampler(BaseSamplingElement, sampler_name="csv_sampler"):

    def __init__(self, filename, counter=0):
        """
            Expects dict of var names, and their corresponding distributions
        """
        self.data = []
        self.filename = filename
        self.counter = counter
        try:
            with open(filename, 'r') as fd:
                reader = csv.DictReader(fd)
                for row in reader:
                    self.data.append(row)
        except FileNotFoundError:
            raise RuntimeError("CSV file you specified ({}) does not exist".format(filename))

    def is_finite(self):
        return True

    def n_samples(self):
        """Returns the number of samples in this sampler.
        Returns
        -------
        if the user specifies maximum number of samples than return that, otherwise - error
        """
        return len(self.data)

    def __next__(self):
        try:
            return self.data[self.counter]
        finally:
            if self.counter < self.n_samples():
                self.counter += 1
            else:
                raise StopIteration



from .base import BaseSamplingElement, Vary
import chaospy as cp
import numpy as np
import pickle
from itertools import product, chain
import logging

__author__ = "Wouter Edeling"
__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class SCSampler(BaseSamplingElement, sampler_name="sc_sampler"):
    """
    Stochastic Collocation sampler
    """

    def __init__(self,
                 vary=None,
                 polynomial_order=4,
                 quadrature_rule="G",
                 count=0,
                 growth=False,
                 sparse=False,
                 midpoint_level1=True,
                 dimension_adaptive=False):
        """
        Create the sampler for the Stochastic Collocation method.

        Parameters
        ----------
        vary: dict or None
            keys = parameters to be sampled, values = distributions.
        polynomial_order : int, optional
            The polynomial order, default is 4.

        quadrature_rule : char, optional
            The quadrature method, default is Gaussian "G".

        growth: bool, optional
             Sets the growth rule to exponential for Clenshaw Curtis quadrature,
             which makes it nested, and therefore more efficient for sparse grids.
             Default is False.

        sparse : bool, optional
            If True use sparse grid instead of normal tensor product grid,
            default is False.
        """

        self.vary = Vary(vary)
        self.quadrature_rule = quadrature_rule

        # List of the probability distributions of uncertain parameters
        params_distribution = list(self.vary.get_values())
        # N = number of uncertain parameters
        self.N = len(params_distribution)

        logging.debug("param dist {}".format(params_distribution))

        # Multivariate distribution
        self.joint_dist = cp.J(*params_distribution)

        # The quadrature information: order, rule and sparsity
        if isinstance(polynomial_order, int):
            logging.debug('Received integer polynomial order, assuming isotropic grid')
            self.polynomial_order = [polynomial_order for i in range(self.N)]
        else:
            self.polynomial_order = polynomial_order

        self.quad_rule = quadrature_rule
        self.sparse = sparse
        # determines how many points the 1st level of a sparse grid will have.
        # If midpoint_level1 = True, order 0 quadrature will be generated
        self.midpoint_level1 = midpoint_level1
        # determines wether to use an insotropic sparse grid, or to adapt
        # the levels in the sparse grid based on a hierachical error measure
        self.dimension_adaptive = dimension_adaptive
        self.nadaptations = 0
        self.quad_sparse = sparse
        self.growth = growth
        self.params_distribution = params_distribution
        self.check_max_quad_level()

        # determine if a nested sparse grid is used
        if self.sparse is True and self.growth is True and \
           (self.quad_rule == "C" or self.quad_rule == "clenshaw_curtis"):
            self.nested = True
        elif self.sparse is True and self.growth is False and self.quad_rule == "gauss_patterson":
            self.nested = True
        elif self.sparse is True and self.growth is True and self.quad_rule == "newton_cotes":
            self.nested = True
        else:
            self.nested = False

        # L = level of (sparse) grid
        self.L = np.max(self.polynomial_order)

        # compute the 1D collocation points (and quad weights)
        self.compute_1D_points_weights(self.L, self.N)

        # compute N-dimensional collocation points
        if not self.sparse:

            # generate collocation grid locally
            l_norm = np.array([self.polynomial_order])
            self.xi_d = self.generate_grid(l_norm)

        # sparse grid = a linear combination of tensor products of 1D rules
        # of different order. Use chaospy to compute these 1D quadrature rules
        else:

            # simplex set of multi indices
            multi_idx = self.compute_sparse_multi_idx(self.L, self.N)

            # create sparse grid of dimension N and level q using the 1d
            #rules in self.xi_1d
            self.xi_d = self.generate_grid(multi_idx)

        self._n_samples = self.xi_d.shape[0]

        self.count = 0

    @property
    def analysis_class(self):
        """Return a corresponding analysis class.
        """
        from easyvvuq.analysis import SCAnalysis
        return SCAnalysis

    def compute_1D_points_weights(self, L, N):
        """
        Computes 1D collocation points and quad weights,
        and stores this in self.xi_1d, self.wi_1d.

        Parameters
        ----------
        L : (int) the max level of the (sparse) grid
        N : (int) the number of uncertain parameters

        Returns
        -------
        None.

        """
        # for every dimension (parameter), create a hierachy of 1D
        # quadrature rules of increasing order
        self.xi_1d = [{} for n in range(N)]
        self.wi_1d = [{} for n in range(N)]

        if self.sparse:

            # if level one of the sparse grid is a midpoint rule, generate
            # the quadrature with order 0 (1 quad point). Else set order at
            # level 1 to 1
            if self.midpoint_level1:
                j = 0
            else:
                j = 1

            for n in range(N):
                # check if input is discrete uniform, in which case the
                # rule and growth flag must be modified
                if isinstance(self.params_distribution[n], cp.DiscreteUniform):
                    rule = "discrete"
                else:
                    rule = self.quad_rule

                for i in range(L):
                    xi_i, wi_i = cp.generate_quadrature(i + j,
                                                        self.params_distribution[n],
                                                        rule=rule,
                                                        growth=self.growth)
                    self.xi_1d[n][i + 1] = xi_i[0]
                    self.wi_1d[n][i + 1] = wi_i
        else:
            for n in range(N):
                # check if input is discrete uniform, in which case the
                # rule flag must be modified
                if isinstance(self.params_distribution[n], cp.DiscreteUniform):
                    rule = "discrete"
                else:
                    rule = self.quad_rule

                xi_i, wi_i = cp.generate_quadrature(self.polynomial_order[n],
                                                    self.params_distribution[n],
                                                    rule=rule,
                                                    growth=self.growth)

                self.xi_1d[n][self.polynomial_order[n]] = xi_i[0]
                self.wi_1d[n][self.polynomial_order[n]] = wi_i

    def check_max_quad_level(self):
        """

        If a discrete variable is specified, there is the possibility of
        non unique collocation points if the quadrature order is high enough.
        This subroutine prevents that.

        NOTE: Only detects cp.DiscreteUniform thus far

        The max quad orders are stores in self.max_quad_order

        Returns
        -------
        None

        """
        # assume no maximum by default
        self.max_level = np.ones(self.N) * 1000
        for n in range(self.N):

            # if a discrete uniform is specified check max order
            if isinstance(self.params_distribution[n], cp.DiscreteUniform):

                # if level one of the sparse grid is a midpoint rule, generate
                # the quadrature with order 0 (1 quad point). Else set order at
                # level 1 to 1
                if self.midpoint_level1:
                    j = 0
                else:
                    j = 1

                number_of_points = 0
                for order in range(1000):
                    xi_i, wi_i = cp.generate_quadrature(order + j,
                                                        self.params_distribution[n],
                                                        growth=self.growth)
                    # if the quadrature points no longer grow with the quad order,
                    # then the max order has been reached
                    if xi_i.size == number_of_points:
                        break
                    number_of_points = xi_i.size

                logging.debug("Input %d is discrete, setting max quadrature order to %d"
                              % (n, order - 1))
                # level 1 = order 0 etc
                self.max_level[n] = order

    def next_level_sparse_grid(self):
        """
        Adds the points of the next level for isotropic hierarchical sparse grids.

        Returns
        -------
        None.

        """

        if self.nested is False:
            logging.debug('Only works for nested sparse grids')
            return

        # update level of sparse grid
        L = np.max(self.polynomial_order) + 1
        self.L = L
        self.polynomial_order = [p + 1 for p in self.polynomial_order]

        logging.debug('Moving grid from level %d to level %d' % (L - 1, L))

        # compute all multi indices
        multi_idx = self.compute_sparse_multi_idx(L, self.N)

        # find only the indices of the new level (|l| = L + N - 1)
        new = np.where(np.sum(multi_idx, axis=1) == L + self.N - 1)[0]

        # update the 1D points and weights
        self.compute_1D_points_weights(L, self.N)

        # generate the new N-dimensional collocation points
        new_grid = self.generate_grid(multi_idx[new])

        # find the new points unique to the new grid
        new_points = setdiff2d(new_grid, self.xi_d)

        logging.debug('%d new points added' % new_points.shape[0])

        # update the number of samples
        self._n_samples += new_points.shape[0]

        # update the N-dimensional sparse grid
        self.xi_d = np.concatenate((self.xi_d, new_points))

    def look_ahead(self, current_multi_idx):
        """
        The look-ahead step in dimension-adaptive sparse grid sampling. Allows
        for anisotropic sampling plans.

        Computes the admissible forward neighbors with respect to the current level
        multi-indices. The admissible level indices l are added to self.admissible_idx.
        The code will be evaluated next iteration at the new collocation points
        corresponding to the levels in admissble_idx.

        Source: Gerstner, Griebel, "Numerical integration using sparse grids"

        Parameters
        ----------
        current_multi_idx : array of the levels in the current iteration
        of the sparse grid.

        Returns
        -------
        None.

        """
        if not self.dimension_adaptive:
            logging.debug('Dimension adaptivity is not selected')
            return

        # compute all forward neighbors for every l in current_multi_idx
        forward_neighbor = []
        e_n = np.eye(self.N, dtype=int)
        for l in current_multi_idx:
            for n in range(self.N):
                # the forward neighbor is l plus a unit vector
                forward_neighbor.append(l + e_n[n])

        # remove duplicates
        forward_neighbor = np.unique(np.array(forward_neighbor), axis=0)
        # remove those which are already in the grid
        forward_neighbor = setdiff2d(forward_neighbor, current_multi_idx)
        # make sure the final candidates are admissible (all backward neighbors
        # must be in the current multi indices)
        logging.debug('Computing admissible levels...')
        admissible_idx = []
        for l in forward_neighbor:
            admissible = True
            for n in range(self.N):
                backward_neighbor = l - e_n[n]
                # find backward_neighbor in current_multi_idx
                idx = np.where((backward_neighbor == current_multi_idx).all(axis=1))[0]
                # if backward neighbor is not in the current index set
                # and it is 'on the interior' (contains no 0): not admissible
                if idx.size == 0 and backward_neighbor[n] != 0:
                    admissible = False
                    break
            # if all backward neighbors are in the current index set: l is admissible
            if admissible:
                admissible_idx.append(l)
        logging.debug('done')

        self.admissible_idx = np.array(admissible_idx)
        # make sure that all entries of each index are <= the max quadrature order
        # The max quad order can be low for discrete input variables
        idx = np.where((self.admissible_idx <= self.max_level).all(axis=1))[0]
        self.admissible_idx = self.admissible_idx[idx]
        logging.debug('Admissible multi-indices:\n', self.admissible_idx)

        # determine the maximum level L of the new index set L = |l| - N + 1
        # self.L = np.max(np.sum(self.admissible_idx, axis=1) - self.N + 1)
        self.L = np.max(self.admissible_idx)
        # recompute the 1D weights and collocation points
        self.compute_1D_points_weights(self.L, self.N)
        # compute collocation grid based on the admissible level indices
        admissible_grid = self.generate_grid(self.admissible_idx)
        # remove collocation points which have already been computed
        new_points = setdiff2d(admissible_grid, self.xi_d)

        logging.debug('%d new points added' % new_points.shape[0])

        # keep track of the number of points added per iteration
        if not hasattr(self, 'n_new_points'):
            self.n_new_points = []
        self.n_new_points.append(new_points.shape[0])

        # update the number of samples
        self._n_samples += new_points.shape[0]

        # update the N-dimensional sparse grid if unsampled points are added
        if new_points.shape[0] > 0:
            self.xi_d = np.concatenate((self.xi_d, new_points))

        # count the number of times the dimensions were adapted
        self.nadaptations += 1

    def is_finite(self):
        return True

    @property
    def n_samples(self):
        """
        Number of samples (Ns) of SC method.
        - When using tensor quadrature: Ns = (p + 1)**d
        - When using sparid: Ns = bigO((p + 1)*log(p + 1)**(d-1))
        Where: p is the polynomial degree and d is the number of
        uncertain parameters.

        Ref: Eck et al. 'A guide to uncertainty quantification and
        sensitivity analysis for cardiovascular applications' [2016].
        """
        return self._n_samples

    # SC collocations points are not random, generate_runs simply returns
    # one collocation point from the tensor product after the other
    def __next__(self):
        if self.count < self._n_samples:
            run_dict = {}
            i_par = 0
            for param_name in self.vary.get_keys():
                run_dict[param_name] = self.xi_d[self.count][i_par]
                i_par += 1
            self.count += 1
            return run_dict
        else:
            raise StopIteration

    def save_state(self, filename):
        logging.debug("Saving sampler state to %s" % filename)
        file = open(filename, 'wb')
        pickle.dump(self.__dict__, file)
        file.close()

    def load_state(self, filename):
        logging.debug("Loading sampler state from %s" % filename)
        file = open(filename, 'rb')
        self.__dict__ = pickle.load(file)
        file.close()

    """
    =========================
    (SPARSE) GRID SUBROUTINES
    =========================
    """

    def generate_grid(self, l_norm):

        dimensions = range(self.N)
        H_L_N = []
        # loop over all multi indices i
        for l in l_norm:
            # compute the tensor product of nodes indexed by i
            X_l = [self.xi_1d[n][l[n]] for n in dimensions]
            H_L_N.append(list(product(*X_l)))
        # flatten the list of lists
        H_L_N = np.array(list(chain(*H_L_N)))
        # return unique nodes
        return np.unique(H_L_N, axis=0)

    def compute_sparse_multi_idx(self, L, N):
        """
        computes all N dimensional multi-indices l = (l1,...,lN) such that
        |l| <= L + N - 1, i.e. a simplex set:
        3    *
        2    *    *          (L=3 and N=2)
        1    *    *    *
             1    2    3
        Here |l| is the internal sum of i (l1+...+lN)
        """
        P = np.array(list(product(range(1, L + 1), repeat=N)))
        multi_idx = P[np.where(np.sum(P, axis=1) <= L + N - 1)[0]]
        return multi_idx


def setdiff2d(X, Y):
    """
    Computes the difference of two 2D arrays X and Y

    Parameters
    ----------
    X : 2D numpy array
    Y : 2D numpy array

    Returns
    -------
    The difference X \ Y as a 2D array

    """
    diff = set(map(tuple, X)) - set(map(tuple, Y))
    return np.array(list(diff))



"""
Summary
-------
This module provides classes based on RandomSampler but modified in such a way
that the output of the sampler is not random but is meant to be used in place
of uniformly random number sequences. Usually this is used to cover the sampling
space more "evenly" than a uniform random distribution would. Two methods are
implemented:

https://en.wikipedia.org/wiki/Latin_hypercube_sampling
https://en.wikipedia.org/wiki/Halton_sequence

"""


from .random import RandomSampler
import numpy as np
import chaospy as cp


class LHCSampler(RandomSampler, sampler_name='lhc_sampler'):
    def __next__(self):
        if self.is_finite():
            if self.count >= self.max_num:
                raise StopIteration

        run_dict = {}
        for param_name, dist in self.vary.get_items():
            run_dict[param_name] = dist.sample(1, rule='L')[0]

        self.count += 1
        return run_dict


class HaltonSampler(RandomSampler, sampler_name='halton_sampler'):
    def __next__(self):
        if self.is_finite():
            if self.count >= self.max_num:
                raise StopIteration

        run_dict = {}
        for param_name, dist in self.vary.get_items():
            run_dict[param_name] = dist.sample(1, rule='H')[0]

        self.count += 1
        return run_dict



"""EasyVVUQ Campaign

This module contains the Campaign class that is used to coordinate all
EasyVVUQ workflows.
"""
import os
import logging
import tempfile
import json
import easyvvuq
from concurrent.futures import ProcessPoolExecutor
from easyvvuq import ParamsSpecification, constants
from easyvvuq.constants import default_campaign_prefix, Status
from easyvvuq.data_structs import RunInfo, CampaignInfo, AppInfo
from easyvvuq.sampling import BaseSamplingElement
from easyvvuq.actions import ActionPool
import easyvvuq.db.sql as db

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


logger = logging.getLogger(__name__)


class Campaign:
    """Campaigns organise the dataflow in EasyVVUQ workflows.

    The Campaign functions as as state machine for the VVUQ workflows. It uses a
    database (CampaignDB) to store information on both the target application
    and the VVUQ algorithms being employed. It also collects data from the simulations
    and can be used to store and resume your state.

    Notes
    -----

    Multiple campaigns can be combined in a CampaignDB. Hence the particular
    campaign we are currently working on will be specified using `campaign_id`.

    Parameters
    ----------
    name : :obj:`str`, optional
    params : dict
        Description of the parameters to associate with the application.
    db_type : str, default="sql"
        Type of database to use for CampaignDB.
    db_location : :obj:`str`, optional
        Location of the underlying campaign database - either a path or
        acceptable URI for SQLAlchemy.
    work_dir : :obj:`str`, optional, default='./'
        Path to working directory - used to store campaign directory.
    state_file : :obj:`str`, optional
        Path to serialised state - used to initialise the Campaign.
    change_to_state : bool, optional, default=False
        Should we change to the directory containing any specified `state_file`
        in order to make relative paths work.
    verify_all_runs: bool, optional, default=True
        Check all new runs being added for unrecognised params (not defined for the currently set
        app), values lying within defined physical range, type checking etc. This should normally
        always be set to True, but in cases where the performance is too degraded, the checks can
        be disabled by setting to False.

    Attributes
    ----------
    campaign_name : str or None
        Name for the campaign/workflow.
    _campaign_dir: str or None
        Path to the directory campaign uses for local storage (runs inputs etc)
    db_location : str or None
        Location of the underlying campaign database - either a path or
        acceptable URI for SQLAlchemy.
    db_type : str or None
        Type of CampaignDB ("sql").
    _log: list
        The log of all elements that have been applied, with information about
        their application
    campaign_id : int
        ID number for the current campaign in the CampaignDB.
    campaign_db: easyvvuq.db.BaseCampaignDB
        A campaign database object
    last_analysis:
        The result of the most recent analysis carried out on this campaign
    _active_app: dict
        Info about currently set app
    _active_app_name: str
        Name of currently set app
    _active_sampler_id: int
        The database id of the currently set Sampler object

    """

    def __init__(
            self,
            name=None,
            params=None,
            actions=None,
            db_type="sql",
            db_location=None,
            work_dir="./",
            state_file=None,
            change_to_state=False,
            verify_all_runs=True
    ):

        self.work_dir = os.path.realpath(os.path.expanduser(work_dir))
        self.verify_all_runs = verify_all_runs

        self.campaign_name = name
        self._campaign_dir = None
        self.db_location = db_location
        self.db_type = db_type

        self.campaign_id = None
        self.campaign_db = None

        self.last_analysis = None

        self._active_app = None
        self._active_app_name = None
        self._active_app_actions = None

        self._active_sampler = None
        self._active_sampler_id = None

        # Load campaign from state_file, if provided. Else make a fresh new
        # campaign with a new campaign database
        if state_file is not None:
            self._state_dir = self.init_from_state_file(
                state_file)
            if change_to_state:
                os.chdir(self._state_dir)
        else:
            self.init_fresh(name, db_type, db_location, self.work_dir)
            self._state_dir = None

        if (params is not None) and (actions is not None):
            self.add_app(name=name, params=params, actions=actions)

    @property
    def campaign_dir(self):
        """Get the path in which to load/save files related to the campaign.

        Returns
        -------
        str
            Path to the campaign directory - given as a subdirectory of the
            working directory.
        """

        return os.path.join(self.work_dir, self._campaign_dir)

    def init_fresh(self, name, db_type='sql',
                   db_location=None, work_dir='.'):
        """
        Initialise a new campaign - create database and output directory
        (`campaign_dir`, a uniquely named directory in `work_dir`).

        Parameters
        ----------
        name : str
            Campaign name.
        db_type : str
            Database type - current options are 'sql'.
        db_location : str or None
            Path in which to create campaign database - defaults to None which
            results in the database being placed in `campaign_dir` with a
            default name.
        work_dir : str
            Path in which to create the `campaign_dir`.

        Returns
        -------

        """

        # Create temp dir for campaign
        campaign_prefix = default_campaign_prefix
        if name is not None:
            campaign_prefix = name

        campaign_dir = tempfile.mkdtemp(prefix=campaign_prefix, dir=work_dir)

        self._campaign_dir = os.path.relpath(campaign_dir, start=work_dir)

        self.db_location = db_location
        self.db_type = db_type

        if self.db_type == 'sql':
            from .db.sql import CampaignDB
            if self.db_location is None:
                self.db_location = "sqlite:///" + self.campaign_dir + "/campaign.db"
        else:
            message = (f"Invalid 'db_type' {db_type}. Supported types are "
                       f"'sql'.")
            logger.critical(message)
            raise RuntimeError(message)

        info = CampaignInfo(
            name=name,
            campaign_dir_prefix=default_campaign_prefix,
            easyvvuq_version=easyvvuq.__version__,
            campaign_dir=self.campaign_dir)
        self.campaign_db = CampaignDB(location=self.db_location,
                                      new_campaign=True,
                                      name=name, info=info)

        # Record the campaign's name and its associated ID in the database
        self.campaign_name = name
        self.campaign_id = self.campaign_db.get_campaign_id(self.campaign_name)

    def init_from_state_file(self, state_file):
        """Load campaign state from file.

        Parameters
        ----------
        state_file : str
            Path to the file containing the campaign state.

        Returns
        -------

        """

        full_state_path = os.path.realpath(os.path.expanduser(state_file))

        if not os.path.isfile(full_state_path):
            if not os.path.exists(full_state_path):
                msg = (f"Unable to open state file (no file exists): "
                       f"{state_file}")
            else:
                msg = (f"Unable to open state file (path is not a file): "
                       f"{state_file}")

            logger.error(msg)
            raise RuntimeError(msg)

        state_dir = os.path.dirname(full_state_path)

        logger.info(f"Loading campaign from state file '{full_state_path}'")
        self.load_state(full_state_path)

        if self.db_type == 'sql':
            from .db.sql import CampaignDB
        else:
            message = (f"Invalid 'db_type' {self.db_type}. Supported types "
                       f"are 'sql'.")
            logger.critical(message)
            raise RuntimeError(message)

        logger.info(f"Opening session with CampaignDB at {self.db_location}")
        self.campaign_db = CampaignDB(location=self.db_location,
                                      new_campaign=False,
                                      name=self.campaign_name)
        campaign_db = self.campaign_db
        self.campaign_id = campaign_db.get_campaign_id(self.campaign_name)

        # Resurrect the sampler
        self._active_sampler_id = campaign_db.get_sampler_id(self.campaign_id)
        self._active_sampler = campaign_db.resurrect_sampler(self._active_sampler_id)

        self.set_app(self._active_app_name)

        return state_dir

    def save_state(self, state_filename):
        """Save the current Campaign state to file in JSON format
        Parameters
        ----------
        state_filename : str
            Name of file in which to save the state
        Returns
        -------
        """

        output_json = {
            "db_location": self.db_location,
            "db_type": self.db_type,
            "active_app": self._active_app_name,
            "campaign_name": self.campaign_name,
            "campaign_dir": self._campaign_dir,
        }
        with open(state_filename, "w") as outfile:
            json.dump(output_json, outfile, indent=4)

    def load_state(self, state_filename):
        """Load up a Campaign state from the specified JSON file

        Parameters
        ----------
        state_filename : str
            Name of file from which to load the state

        Returns
        -------

        """

        with open(state_filename, "r") as infile:
            input_json = json.load(infile)
        self.db_location = input_json["db_location"]
        self.db_type = input_json["db_type"]
        self._active_app_name = input_json["active_app"]
        self.campaign_name = input_json["campaign_name"]
        self._campaign_dir = input_json["campaign_dir"]

        if not os.path.exists(self.campaign_dir):
            message = (f"Campaign directory in state_file {state_filename}"
                       f" ({self.campaign_dir}) does not exist.")
            logger.critical(message)
            raise RuntimeError(message)

    def add_app(self, name=None, params=None, actions=None, set_active=True):
        """Add an application to the CampaignDB.

        Parameters
        ----------
        name : str
            Name of the application.
        params : dict
            Description of the parameters to associate with the application.
        decoderspec : dict
            A Cerberus validation dictionary for the output of the decoder
            if left None, will not be used.
        encoder : :obj:`easyvvuq.encoders.base.BaseEncoder`
            Encoder element to convert parameters into application run inputs.
        decoder : :obj:`easyvvuq.decoders.base.BaseDecoder`
            Decoder element to convert application run output into data for
            VVUQ analysis.
        set_active: bool
            Should the added app be set to be the currently active app?

        Returns
        -------

        """

        # Verify input parameters dict
        paramsspec = ParamsSpecification(params, appname=name)

        # validate application input
        app = AppInfo(
            name=name,
            paramsspec=paramsspec,
            actions=actions,
        )

        self.campaign_db.add_app(app)
        if set_active:
            self.set_app(app.name)

    def set_app(self, app_name):
        """Set active app for the campaign.

        Application information is retrieved from `self.campaign_db`.

        Parameters
        ----------
        app_name: str or None
            Name of selected app, if `None` given then first app will be
            selected.

        Returns
        -------

        """

        self._active_app_name = app_name
        self._active_app = self.campaign_db.app(name=app_name)

        # Resurrect the app encoder, decoder and collation elements
        self._active_app_actions = self.campaign_db.resurrect_app(app_name)

    def set_sampler(self, sampler, update=False):
        """Set active sampler.

        Parameters
        ----------
        sampler : easyvvuq.sampling.base.BaseSamplingElement
            Sampler that will be used to create runs for the current campaign.
        update : bool
            If set to True it will not add the sampler to the database, just change it as the active sampler.

        Returns
        -------

        """
        if not isinstance(sampler, BaseSamplingElement):
            msg = "set_sampler() must be passed a sampling element"
            logging.error(msg)
            raise Exception(msg)

        self._active_sampler = sampler
        if not update:
            self._active_sampler_id = self.campaign_db.add_sampler(sampler)
            sampler.sampler_id = self._active_sampler_id
        self._active_sampler_id = self._active_sampler.sampler_id
        self.campaign_db.set_sampler(self.campaign_id, self._active_sampler.sampler_id)

    def add_runs(self, runs, mark_invalid=False):
        """Add a new run to the queue.

        Parameters
        ----------
        runs : list of dicts
            Each dict defines the value of each model parameter listed in
            `self.params_info` for a run to be added to `self.runs`
        mark_invalid : bool
            Will mark runs that fail verification as invalid (but will not raise an exception)

        Returns
        -------

        """

        if self._active_app is None:
            msg = ("No app is currently set for this campaign. "
                   "Use set_app('name_of_app').")
            logging.error(msg)
            raise Exception(msg)

        app_default_params = self._active_app["params"]

        run_info_list = []
        for new_run in runs:

            if new_run is None:
                msg = ("add_run() was passed new_run of type None. Bad sampler?")
                logging.error(msg)
                raise Exception(msg)

            # Verify and complete run with missing/default param values
            status = Status.NEW
            try:
                new_run = app_default_params.process_run(new_run, verify=self.verify_all_runs)
            except RuntimeError:
                if mark_invalid:
                    new_run = app_default_params.process_run(new_run, verify=False)
                    status = Status.INVALID
                else:
                    raise

            # Add to run queue
            run_info = RunInfo(app=self._active_app['id'],
                               params=new_run,
                               sample=self._active_sampler_id,
                               campaign=self.campaign_id,
                               status=status)

            run_info_list.append(run_info)

        self.campaign_db.add_runs(run_info_list, iteration=self._active_sampler.iteration)

    def add_default_run(self):
        """
        Add a single new run to the queue, using only default values for
        all parameters.
        """

        new_run = {}
        self.add_runs([new_run])

    def draw_samples(self, num_samples=0, replicas=1, mark_invalid=False):
        """Draws `num_samples` sets of parameters from the currently set
        sampler, resulting in `num_samples` * `replicas` new runs added to the
        runs list. If `num_samples` is 0 (its default value) then
        this method draws ALL samples from the sampler, until exhaustion (this
        will fail if the sampler is not finite).

        Notes
        -----
        Do NOT use this in cases where you need 'replicas' with a different
        random seed. In those cases need to sample from a uniform_integer
        distribution and filter at analysis step.

        Parameters
        ----------
        num_samples : int
                Number of samples to draw from the active sampling element.
                By default is 0 (draw ALL samples)
        replicas : int
                Number of replica runs to create with each set of parameters.
                Default is 1 - so only a single run added for each set of
                parameters.
        mark_invalid : bool
           If True will mark runs that go outside valid parameter range as INVALID.
           This is useful for MCMC style methods where you want those runs to evaluate
           to low probabilities.

        Returns
        -------

        """

        # Make sure `num_samples` is not 0 for an infinite generator
        # (this would add runs forever...)
        if not self._active_sampler.is_finite() and num_samples <= 0:
            msg = (f"Sampling_element '{self._active_sampler.element_name()}' "
                   f"is an infinite generator, therefore a finite number of "
                   f"draws (n > 0) must be specified.")
            raise RuntimeError(msg)

        if replicas < 1:
            msg = "Number of replicas ({replicas}) must be at least 1"
            raise RuntimeError(msg)

        num_added = 0
        new_runs = []
        for new_run in self._active_sampler:

            list_of_runs = [new_run for i in range(replicas)]
            new_runs += list_of_runs

            num_added += 1

            if num_samples != 0 and num_added >= num_samples:
                break

        self.add_runs(new_runs, mark_invalid)

        # Write sampler's new state to database
        self.campaign_db.update_sampler(self._active_sampler_id, self._active_sampler)
        return new_runs

    def list_runs(self, sampler=None, campaign=None, status=None):
        """Get list of runs in the CampaignDB.

        Returns
        -------
            list of runs

        """
        return list(self.campaign_db.runs(sampler=sampler, campaign=campaign, status=status))

    def scan_completed(self, *args, **kwargs):
        """
        Check campaign database for completed runs (defined as runs with COLLATED status)

        Returns
        -------
            list of runs

        """
        return self.list_runs(status=Status.COLLATED)

    def all_complete(self):
        """
        Check if all runs have reported having output generated by
        a completed simulation.

        Returns
        -------
            list of runs

        """

        num = self.campaign_db.get_num_runs(not_status=Status.COLLATED)
        if num == 0:
            return True
        return False

    def get_campaign_runs_dir(self):
        """Get the runs directory from the CampaignDB.

        Returns
        -------
        str
            Path in which the runs information will be written.

        """
        return self.campaign_db.runs_dir(self.campaign_name)

    def relocate(self, campaign_dir):
        """Relocate the campaign by specifying a new path where campaign is located.

        Parameters
        ----------
        new_path: str
            new runs directory
        """
        if not os.path.exists(campaign_dir):
            raise RuntimeError("specified directory does not exist: {}".format(campaign_dir))
        self.campaign_db.relocate(campaign_dir, self.campaign_name)

    def execute(self, nsamples=0, pool=None, mark_invalid=False, sequential=False):
        """This will draw samples and execute the action on those samples.
        
        Parameters
        ----------
        nsamples : int
            Number of samples to draw. For infinite samplers or when you want to process samples in batches.
        pool : Pool
            A pool object to be used when processing runs (e.g. instance of ThreadPoolExecutor or 
            ProcessPoolExecutor).
        mark_invalid : bool
            Mark runs that go outside the specified input parameter range as INVALID.
        sequential : bool
            Whether to process samples sequentially (sometimes more efficient).
        """
        self.draw_samples(nsamples, mark_invalid=mark_invalid)
        action_pool = self.apply_for_each_sample(
            self._active_app_actions, sequential=sequential)
        return action_pool.start(pool=pool)

    def apply_for_each_sample(self, action, sequential=False):
        """
        For each run in this Campaign's run list, apply the specified action
        (an object of type Action)

        Parameters
        ----------
        action : the action to be applied to each run directory
            The function to be applied to each run directory. func() will
            be called with the run directory path as its only argument.
        status : Status
            Will apply the action only to those runs whose status is as specified

        Returns
        -------
        action_statuses: ActionStatuses
            An object containing ActionStatus instances to track action execution
        """

        # Loop through all runs in this campaign with status ENCODED, and
        # run the specified action on each run's dir
        def inits():
            for run_id, run_data in self.campaign_db.runs(status=Status.NEW, app_id=self._active_app['id']):
                previous = {}
                previous['run_id'] = run_id
                previous['campaign_dir'] = self._campaign_dir
                previous['run_info'] = run_data
                yield previous
        return ActionPool(self, action, inits=inits(), sequential=sequential)
        

    def iterate(self, nsamples=0, pool=None, mark_invalid=False, sequential=False):
        """This is the equivalent of sample_and_apply for methods that rely on the output of the
        previous sampling stage (primarily MCMC).

        Parameters
        ----------
        nsamples : int
            number of samples to draw
        action : BaseAction
            an action to be executed
        *args : args
            arguments to action
        max_workers : int
            number of actions to be executed at the same time
        mark_invalid : bool
            Mark runs that go outside the specified input parameter range as INVALID.

        Yields
        ------
        action_statuses: ActionStatuses
            An object containing ActionStatus instances to track action execution
        """
        while True:
            self.draw_samples(nsamples, mark_invalid=mark_invalid)
            action_pool = self.apply_for_each_sample(
                self._active_app_actions, sequential=sequential)
            yield action_pool.start(pool=pool)
            result = self.get_collation_result(last_iteration=True)
            invalid = self.get_invalid_runs(last_iteration=True)
            ignored_runs = self._active_sampler.update(result, invalid)
            for run_id in ignored_runs:
                self.campaign_db.session.query(db.RunTable).\
                    filter(db.RunTable.id == int(run_id)).\
                    update({'status': constants.Status.IGNORED})
            self.campaign_db.session.commit()

    def recollate(self):
        """Clears the current collation table, changes all COLLATED status runs
           back to ENCODED, then runs collate() again

        Returns
        -------

        """

        collated_run_ids = list(self.campaign_db.run_ids(status=Status.COLLATED))
        self.campaign_db.set_run_statuses(collated_run_ids, Status.ENCODED)
        self.collate()

    def get_collation_result(self, last_iteration=False):
        """
        Return dataframe containing all collated results

        Parameters
        ----------
        last_collation : bool
            Will only return the result of the last collation.

        Returns
        -------
            pandas dataframe

        """
        if last_iteration:
            iteration = self._active_sampler.iteration - 1
        else:
            iteration = -1
        return self.campaign_db.get_results(self._active_app['name'], self._active_sampler_id,
                                            status=constants.Status.COLLATED, iteration=iteration)

    def get_invalid_runs(self, last_iteration=False):
        """
        Return dataframe containing all results marked as INVALID.

        Parameters
        ----------
        last_collation : bool
            Will only return the result of the last collation.

        Returns
        -------
            pandas DataFrame
        """
        if last_iteration:
            iteration = self._active_sampler.iteration - 1
        else:
            iteration = -1
        return self.campaign_db.get_results(self._active_app['name'], self._active_sampler_id,
                                            status=constants.Status.INVALID, iteration=iteration)

    def apply_analysis(self, analysis):
        """Run the `analysis` element on the output of the last run collation.

        Parameters
        ----------
        analysis : `easyvvuq.analysis.base.BaseAnalysisElement`
            Element that performs a VVUQ analysis on a dataframe summary of
            run outputs.

        Returns
        -------

        """

        # Apply analysis element to most recent collation result
        self.last_analysis = analysis.analyse(data_frame=self.get_collation_result())

    def analyse(self, **kwargs):
        """If available will call an appropriate analysis class on the collation result.

        Parameters
        ----------
        **kwargs - dict
           Argument to the analysis class constructor (after sampler).
        """
        collation_result = self.get_collation_result()
        try:
            analysis = self._active_sampler.analysis_class(sampler=self._active_sampler, **kwargs)
            return analysis.analyse(collation_result)
        except NotImplementedError:
            raise RuntimeError("This sampler does not have a corresponding analysis class")

    def get_last_analysis(self):
        """
        Return the output of the most recently run analysis element.

        Returns
        -------

        """
        if self.last_analysis is None:
            logging.warning("No last analysis output available.")

        return self.last_analysis

    def __str__(self):
        """Returns formatted summary of the current Campaign state.
        Enables class to work with standard print() method"""

        return (f"db_location = {self.db_location}\n"
                f"active_sampler_id = {self._active_sampler_id}\n"
                f"campaign_name = {self.campaign_name}\n"
                f"campaign_dir = {self.campaign_dir}\n"
                f"campaign_id = {self.campaign_id}\n")

    def get_active_sampler(self):
        """ Return the active sampler element in use by this campaign.

        Parameters
        ----------

        Returns
        -------
        The sampler currently in use
            easyvvuq.sampling.base.BaseSamplingElement

        """

        return self._active_sampler

    def ignore_runs(self, list_of_run_IDs):
        """ Flags the specified runs to be IGNORED in future collation. Note that
        this does NOT remove previously collated results from the collation table.
        For that you must refresh the collation by running recollate().

        Parameters
        ----------
        list_of_run_IDs: list
            The list of run IDs for the runs that should be set to status IGNORED


        Returns
        -------

        """
        self.campaign_db.set_run_statuses(list_of_run_IDs, Status.IGNORED)

    def rerun(self, list_of_run_IDs):
        """ Sets the status of the specified runs to ENCODED, so that their results
        may be recollated later (presumably after extending, rerunning or otherwise
        modifying the data in the relevant run folder). Note that this method will
        NOT perform any execution - it simply flags the run in EasyVVUQ as being
        uncollated. Actual execution is (as usual) the job of the user or middleware.

        Parameters
        ----------
        list_of_run_IDs: list
            The list of run IDs for the runs that should be set to status ENCODED


        Returns
        -------

        """

        for run_ID in list_of_run_IDs:
            status = self.campaign_db.get_run_status(run_ID)
            if status == Status.NEW:
                msg = (f"Cannot rerun {run_ID} as it has status NEW, and must"
                       f"be encoded before execution.")
                raise RuntimeError(msg)

        self.campaign_db.set_run_statuses(list_of_run_IDs, Status.ENCODED)

    def get_active_app(self):
        """
        Returns a dict of information regarding the application that is currently
        set for this campaign.
        """
        return self._active_app



import os
import easyvvuq as uq
import numpy as np
import chaospy as cp
import pytest
import logging
import pandas as pd
import json
from tests.sc.sobol_model import sobol_g_func
from easyvvuq.analysis.qmc_analysis import QMCAnalysisResults


def exact_sobols_g_func(d=2, a=[0.0, 0.5, 3.0, 9.0, 99.0]):
    # for the Sobol g function, the exact (1st-order)
    # Sobol indices are known analytically
    V_i = np.zeros(d)
    for i in range(d):
        V_i[i] = 1.0 / (3.0 * (1 + a[i])**2)
    V = np.prod(1 + V_i) - 1
    logging.debug('Exact 1st-order Sobol indices: ', V_i / V)
    return V_i / V


@pytest.fixture
def data():
    # fix random seed to make this test deterministic
    np.random.seed(10000000)
    # Create the sampler
    vary = {
        "x1": cp.Uniform(0.0, 1.0),
        "x2": cp.Uniform(0.0, 1.0)
    }
    sampler = uq.sampling.MCSampler(vary, n_mc_samples=100)
    data = {('run_id', 0): [], ('x1', 0): [], ('x2', 0): [], ('f', 0): []}
    for run_id, sample in enumerate(sampler):
        data[('run_id', 0)].append(run_id)
        data[('x1', 0)].append(sample['x1'])
        data[('x2', 0)].append(sample['x2'])
        data[('f', 0)].append(sobol_g_func([sample['x1'], sample['x2']], d=2))
    df = pd.DataFrame(data)
    return sampler, df


@pytest.fixture
def data_vectors():
    np.random.seed(10000000)
    vary = {
        "x1": cp.Uniform(0.0, 1.0),
        "x2": cp.Uniform(0.0, 1.0)
    }
    sampler = uq.sampling.MCSampler(vary, n_mc_samples=100)
    data = {('run_id', 0): [], ('x1', 0): [], ('x2', 0): [],
            ('g', 0): [], ('g', 1): [], ('g', 2): [], ('h', 0): [], ('h', 1): []}
    for run_id, sample in enumerate(sampler):
        data[('run_id', 0)].append(run_id)
        data[('x1', 0)].append(sample['x1'])
        data[('x2', 0)].append(sample['x2'])
        data[('g', 0)].append(sample['x1'])
        data[('g', 1)].append(sample['x2'])
        data[('g', 2)].append(sample['x1'] + sample['x2'])
        data[('h', 0)].append(sample['x1'] * sample['x2'])
        data[('h', 1)].append(sample['x1'] ** sample['x2'])
    df = pd.DataFrame(data)
    return sampler, df


@pytest.fixture
def results(data):
    # Post-processing analysis
    mc_sampler, df = data
    analysis = uq.analysis.QMCAnalysis(sampler=mc_sampler, qoi_cols=['f'])
    results = analysis.analyse(df)
    return results


@pytest.fixture
def results_vectors(data_vectors):
    # Post-processing analysis
    mc_sampler, df = data_vectors
    analysis = uq.analysis.QMCAnalysis(sampler=mc_sampler, qoi_cols=['g', 'h'])
    results = analysis.analyse(df)
    return results


def test_results(results):
    assert(isinstance(results, QMCAnalysisResults))
    sobols_first_x1 = results._get_sobols_first('f', 'x1')
    sobols_first_x2 = results._get_sobols_first('f', 'x2')
    sobols_total_x1 = results._get_sobols_total('f', 'x1')
    sobols_total_x2 = results._get_sobols_total('f', 'x2')
    assert(sobols_first_x1 == pytest.approx(0.55690589, 0.001))
    assert(sobols_first_x2 == pytest.approx(0.20727553, 0.001))
    assert(sobols_total_x1 == pytest.approx(0.81327937, 0.001))
    assert(sobols_total_x2 == pytest.approx(0.38049629, 0.001))


def test_results_conf(results):
    sobols_first_x1_conf = results._get_sobols_first_conf('f', 'x1')
    assert(sobols_first_x1_conf[0] == pytest.approx(0.14387, 0.001))
    assert(sobols_first_x1_conf[1] == pytest.approx(0.894288, 0.001))
    sobols_first_x2_conf = results._get_sobols_first_conf('f', 'x2')
    assert(sobols_first_x2_conf[0] == pytest.approx(-0.110633, 0.001))
    assert(sobols_first_x2_conf[1] == pytest.approx(0.467528, 0.001))
    sobols_total_x1_conf = results._get_sobols_total_conf('f', 'x1')
    assert(sobols_total_x1_conf[0] == pytest.approx(0.613689, 0.001))
    assert(sobols_total_x1_conf[1] == pytest.approx(1.018587, 0.001))
    sobols_total_x2_conf = results._get_sobols_total_conf('f', 'x2')
    assert(sobols_total_x2_conf[0] == pytest.approx(0.243612, 0.001))
    assert(sobols_total_x2_conf[1] == pytest.approx(0.492141, 0.001))


def test_full_results(results):
    assert(results.sobols_first() == {'f': {'x1': 0.5569058947880715, 'x2': 0.20727553481694053}})
    assert(results.sobols_total() == {'f': {'x1': 0.8132793654841785, 'x2': 0.3804962894947435}})


def test_describe(results_vectors):
    assert(
        results_vectors.describe()[
            ('g',
             1)].to_dict() == {
            'mean': 0.4691844466934421,
            'var': 0.08534945020531205,
            'std': 0.29214628220347433})
    assert(
        results_vectors.describe('h')[
            ('h',
             1)].to_dict() == {
            'mean': 0.6873389710989142,
            'var': 0.07501266456861228,
            'std': 0.27388440000958847})
    assert(isinstance(results_vectors.describe('h', 'std'), np.ndarray))



import pytest
from easyvvuq.sampling import CSVSampler


def test_csv_sampler():
    with pytest.raises(RuntimeError):
        CSVSampler('tests/simple_csv/test.cs')
    sampler = CSVSampler('tests/simple_csv/test.csv')
    counter = 0
    for sample in sampler:
        if sample['Step'] == 5:
            assert(sample['Value'] == 25.950662)
        counter += 1
    assert(counter == 10)
    assert(sampler.n_samples() == 10)
    assert(sampler.is_finite())



from concurrent.futures import ThreadPoolExecutor, as_completed
from dask.distributed import Client
import copy

__copyright__ = """

    Copyright 2020 Vytautas Jancauskas

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class ActionPool:
    """A class that tracks statuses of a list of actions.

    Parameters
    ----------
    statuses: list of ActionStatus
        a list of action statuses to track
    poll_sleep_time: int
        a time to sleep for after iterating over all active statuses
        before starting again

    """

    def __init__(self, campaign, actions, inits, sequential=False):
        self.campaign = campaign
        self.actions = actions
        self.inits = inits
        self.sequential = sequential
        self.futures = []
        self.results = []

    def start(self, pool=None):
        """Start the actions.

        Returns
        -------
        A list of Python futures represending action execution.
        """
        if pool is None:
            pool = ThreadPoolExecutor()
        self.pool = pool
        for previous in self.inits:
            previous = copy.copy(previous)
            if self.sequential:
                result = self.actions.start(previous)
                self.results.append(result)
            else:
                future = self.pool.submit(self.actions.start, previous)
                self.futures.append(future)
        return self

    def progress(self):
        """Some basic stats about the action statuses status.

        Returns
        -------
        A dictionary with four keys - 'ready', 'active' and 'finished', 'failed'.
        """
        ready = 0
        running = 0
        done = 0
        failed = 0
        for future in self.futures:
            if future.running():
                running += 1
            elif future.done():
                if not future.result():
                    failed += 1
                else:
                    done += 1
            else:
                ready += 1
        return {'ready': ready, 'active': running, 'finished': done, 'failed': failed}

    def collate(self):
        """A command that will block untill all futures in the pool have finished.
        """
        if isinstance(self.pool, Client):
            self.results = self.pool.gather(self.futures)
        if self.sequential or isinstance(self.pool, Client):
            for result in self.results:
                self.campaign.campaign_db.store_result(result['run_id'], result)
        else:
            for future in as_completed(self.futures):
                result = future.result()
                self.campaign.campaign_db.store_result(result['run_id'], result)
        self.campaign.campaign_db.session.commit()



import numpy as np
import chaospy as cp
import easyvvuq as uq
import pytest


__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__author__ = 'Jalal Lakhlili'
__license__ = "LGPL"


def test_validate_similarity():
    pass


def test_validate_similarity_hellinger():
    validator = uq.comparison.validate.ValidateSimilarityHellinger()
    assert(validator.element_name() == 'validate_similarity_hellinger')
    assert(validator.element_version() == '0.1')
    d1 = cp.Exponential(1)
    d2 = cp.Exponential(2)
    xmin = min(d1.lower[0], d2.lower[0])
    xmax = max(d1.upper[0], d2.upper[0])
    x = np.linspace(xmin, xmax, 100)
    p1 = d1.pdf(x)
    p2 = d2.pdf(x)
    distance = validator.compare(p1, p2)
    err = abs(distance - np.sqrt(1 - 2 * np.sqrt(2) / 3))
    assert err < 1.e-2


def test_validate_similarity_jensen_shannon():
    validator = uq.comparison.validate.ValidateSimilarityJensenShannon()
    assert(validator.element_name() == 'validate_similarity_jensen_shannon')
    assert(validator.element_version() == '0.1')
    d1 = cp.Normal(0, 1)
    d2 = cp.Normal(1, 2)
    xmin = min(d1.lower[0], d2.lower[0])
    xmax = max(d1.upper[0], d2.upper[0])
    x = np.linspace(xmin, xmax, 100)
    p1 = d1.pdf(x)
    p2 = d2.pdf(x)
    distance = validator.compare(p1, p2)
    assert distance >= 0.0 and distance <= 1.0


def test_validate_similarity_wasserstein():
    validator = uq.comparison.validate.ValidateSimilarityWasserstein()
    assert(validator.element_name() == 'validate_similarity_wasserstein')
    assert(validator.element_version() == '0.1')
    d1 = cp.Normal(0, 1)
    d2 = cp.Normal(1, 2)
    xmin = min(d1.lower[0], d2.lower[0])
    xmax = max(d1.upper[0], d2.upper[0])
    x = np.linspace(xmin, xmax, 100)
    p1 = (xmax - xmin) * d1.cdf(x)
    p2 = (xmax - xmin) * d2.cdf(x)
    distance = validator.compare(p1, p2)
    assert distance >= 0.0



"""Provides baseclass for all decoders and dictionary to register all imported
decoders.

Decoders are objects which provide functions to check simulation runs have
completed and parse the output when they have.

Attributes
----------
AVAILABLE_DECODERS : dict
    Registers all imported decoders.
"""
from easyvvuq.base_element import BaseElement
import json

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


# Dict to store all registered decoders (any class which extends
# BaseDecoder is automatically registered as an decoder)
AVAILABLE_DECODERS = {}


class BaseDecoder(BaseElement):
    """Baseclass for all EasyVVUQ decoders.

    Skeleton decoder which establishes the format and provides the basis of our
    contract - <need to define contract>.


    Parameters
    ----------

    Attributes
    ----------

    """

    def __init_subclass__(cls, decoder_name, **kwargs):
        """
        Catch any new decoders (all decoders must inherit from BaseDecoder) and add them
        to the dict of available decoders.
        """
        super().__init_subclass__(**kwargs)

        cls.decoder_name = decoder_name

        # Register new decoder
        AVAILABLE_DECODERS[decoder_name] = cls

    def sim_complete(self, run_info=None):
        """
        Check whether the simulation specified by `run_info` has completed and
        produced results.

        Parameters
        ----------
        run_info: dict or None
            Information defining the run to check.

        Returns
        -------

        """
        raise NotImplementedError

    def parse_sim_output(self, run_info=None):
        """

        Parameters
        ----------
        run_info: dict or None
            Information defining the run for which we want to parse the output.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.

        Returns
        -------

        """
        raise NotImplementedError



import os
import logging
from easyvvuq import OutputType
from .base import BaseDecoder
import json
import numpy as np

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


logger = logging.Logger(__name__)


class JSONDecoder(BaseDecoder, decoder_name="json"):

    def __init__(self, target_filename=None, output_columns=None):

        if target_filename is None:
            msg = (
                f"target_filename must be set for {self.decoder_name.upper()}Decoder. "
                f"This should be the name of the output file this decoder "
                f"acts on."
            )
            logging.error(msg)
            raise RuntimeError(msg)

        if output_columns is None:
            msg = (
                f"output_columns must be specified for "
                f"{self.decoder_name.upper()}Decoder. This should "
                f"be the names of the output fields this decoder extracts "
                f"from the target json file. For nested values use arrays "
                f"e.g. ['root', 'node', 'field'] where field is the leaf "
                f"that contains the value you need."
            )
            logging.error(msg)
            raise RuntimeError(msg)

        if len(output_columns) == 0:
            msg = "output_columns cannot be empty."
            logger.error(msg)
            raise RuntimeError(msg)

        self.target_filename = target_filename
        self.output_columns = output_columns

        self.output_type = OutputType('sample')

    @staticmethod
    def _get_output_path(run_info=None, outfile=None):

        run_path = run_info['run_dir']

        if not os.path.isdir(run_path):
            raise RuntimeError(f"Run directory does not exist: {run_path}")

        return os.path.join(run_path, outfile)

    def sim_complete(self, run_info=None):

        out_path = self._get_output_path(run_info, self.target_filename)

        if not os.path.isfile(out_path):
            return False
        else:
            return True

    def parse_sim_output(self, run_info={}):
        def get_value(data, path):
            for node in path:
                data = data[node]
            return data

        out_path = self._get_output_path(run_info, self.target_filename)
        raw_data = self._get_raw_data(out_path)

        data = []
        for col in self.output_columns:
            try:
                if isinstance(col, str):
                    data.append((col, raw_data[col]))
                elif isinstance(col, list):
                    data.append(('.'.join(col), get_value(raw_data, col)))
            except KeyError:
                raise RuntimeError("no such field: {} in this json file".format(col))
        return dict(data)

    def _get_raw_data(self, out_path):
        with open(out_path) as fd:
            return json.load(fd)
